// 32 bit RGBA bitmap functions.
// Written by Nils Liaaen Corneliusen and Sjur Julin.
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
// https://www.ignorantus.com
// https://www.in-a-vision.com

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <stdint.h>

#include <libpng/png.h>

#include "bmp.h"
#include "fontdefs.h"

bmp32 *bmp32_Create( int w, int h, void *data )
{
	bmp32 *bp = malloc( sizeof(bmp32) );
	bp->data = data;
	bp->w = w;
	bp->h = h;
	return bp;
}

bmp32 *bmp32_Alloc( int w, int h )
{
	bmp32 *bp  = malloc( sizeof(bmp32) );
	bp->data   = malloc( w*h*sizeof(uint32_t) );
	bp->w = w;
	bp->h = h;
	bmp32_Clear( bp );
	return bp;
}

bmp32 *bmp32_Duplicate( bmp32 *src )
{
	bmp32 *bp = bmp32_Alloc( src->w, src->h );
	memcpy( bp->data, src->data, src->w*src->h*sizeof(uint32_t) );
	return bp;
}

void bmp32_Clear( bmp32 *dst )
{
	memset( dst->data, 0x00, dst->w*dst->h*sizeof(uint32_t) );
}

void bmp32_ClearAlpha( bmp32 *dst )
{
	for( int y = 0; y < dst->h; y++ ) {
		uint32_t *row = dst->data + y*dst->w;
		for( int x = 0; x < dst->w; x++ ) {
			*row++ = 0xff000000;
		}
	}
}

void bmp32_DrawRect( bmp32 *dst, int ulx, int uly, int brx, int bry, uint32_t col )
{
	for( int y = uly; y <= bry; y++ ) {

		uint32_t *row = dst->data + y*dst->w;

		for( int x = ulx; x <= brx; x++ ) {

			row[x] = col;

		}

	}

}

bmp32 *bmp32_ReadPng( const char *fname, bool flip )
{
	FILE *fd = fopen( fname, "rb" );
	if( fd == NULL ) {
		printf( "ReadPng: Can't open %s!\n", fname );
		return NULL;
	}

	fseek( fd, 0, SEEK_END );
	size_t fileSize = ftell( fd );
	fseek( fd, 0, SEEK_SET );

	if( fileSize == 0 ) {
		printf( "ReadPng: File %s is empty!\n", fname );
		fclose( fd );
		return NULL;
	}

	uint8_t *fileBuffer = (uint8_t *)malloc( fileSize );
	size_t readBytes = fread( fileBuffer, 1, fileSize, fd );

	fclose( fd );

	if( readBytes != fileSize ) {
		printf( "ReadPng: Error reading from %s. Got %zu, expected %zu.\n", fname, readBytes, fileSize );
		free( fileBuffer );
		return NULL;
	}

	bmp32 *dst = bmp32_ReadPngMem( fileBuffer, fileSize, flip );

	if( dst == NULL ) {
		printf( "ReadPng: ReadPngMem failed. File %s\n", fname );
	}

	free( fileBuffer );

	return dst;
}

static void read_mem_fn( png_structp png_ptr, png_bytep outBuf, png_size_t byteCount )
{
	png_voidp io_ptr = png_get_io_ptr( png_ptr );
	if( io_ptr == NULL ) return;

	uint8_t **mem = (uint8_t **)io_ptr;
	memcpy( outBuf, *mem, byteCount );
	*mem += byteCount;
}

bmp32 *bmp32_ReadPngMem( uint8_t *ppng, size_t size, bool flip )
{
	uint8_t *mem = ppng;

	png_structp png_ptr = NULL;
	png_infop info_ptr  = NULL;
	png_infop end_info  = NULL;

	png_byte header[8];

	memcpy( header, mem, 8 ); mem += 8;// read header and check if it is png

	if( !png_check_sig( header, 8 ) ) {
		printf( "ReadPngMem: Invalid PNG header\n" );
		return NULL;
	}

	png_ptr  = png_create_read_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL );
	info_ptr = png_create_info_struct( png_ptr );
	end_info = png_create_info_struct( png_ptr );

	if( png_ptr == NULL || info_ptr == NULL || end_info == NULL ) {
		printf( "ReadPngMem: Could not create some struct\n" );
		png_destroy_read_struct( &png_ptr, &info_ptr, &end_info );
		return NULL;
	}

	png_set_read_fn( png_ptr, (png_voidp)&mem, read_mem_fn );
	png_set_sig_bytes( png_ptr, 8 ); // skip signature that's already checked

	png_read_info( png_ptr, info_ptr );
	png_uint_32 w      = png_get_image_width( png_ptr, info_ptr );
	png_uint_32 h      = png_get_image_height(png_ptr, info_ptr );
	png_byte    ctype  = png_get_color_type(  png_ptr, info_ptr );
	png_byte    chans  = png_get_channels(    png_ptr, info_ptr );
						 png_set_interlace_handling( png_ptr );

	png_read_update_info( png_ptr, info_ptr );

	if( chans != 4 || ctype != PNG_COLOR_TYPE_RGBA ) {
		printf( "ReadPngMem: Illegal channels or color type (%d,%d)\n", chans, ctype );
		png_destroy_read_struct( &png_ptr, &info_ptr, &end_info );
		return NULL;
	}

	bmp32 *dst = bmp32_Alloc( w, h );

	png_bytep *rowptrs = (png_bytep *)malloc( h * sizeof(png_bytep *) );

	for( int i = 0; i < dst->h; i++ )
		rowptrs[i] = (png_byte *)&dst->data[dst->w * (flip ? dst->h-1-i : i)];

	png_read_image( png_ptr, rowptrs );

	png_destroy_read_struct( &png_ptr, &info_ptr, &end_info );

	free( rowptrs );

	return dst;
}

bool bmp32_WritePng( bmp32 *src, const char *fname, bool flip )
{
	png_infop info_ptr  = NULL;
	png_structp png_ptr = NULL;
	png_byte color_type = PNG_COLOR_TYPE_RGBA;

	FILE *fp;
	fp = fopen( fname, "wb" );
	if( fp == NULL ) {
		printf( "WritePng: %s - Open error\n", fname );
		return false;
	}

	png_ptr  = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	info_ptr = png_create_info_struct(png_ptr);

	if( png_ptr == NULL || info_ptr == NULL ) {
		printf( "WritePng: %s - Could not create some struct\n", fname );
		png_destroy_write_struct( &png_ptr, &info_ptr );
		fclose( fp );
		return false;
	}

	if( setjmp( png_jmpbuf( png_ptr ) ) ) {
		png_destroy_write_struct(&png_ptr, &info_ptr );
		fprintf( stderr, "Write: %s - setjmp() failed\n", fname );
		png_destroy_write_struct( &png_ptr, &info_ptr );
		fclose( fp );
		return false;
	}

	png_init_io( png_ptr, fp );

	png_set_IHDR( png_ptr, info_ptr, src->w, src->h,
				  8, color_type, PNG_INTERLACE_NONE,
				  PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT );
	png_set_filter( png_ptr, 0, PNG_ALL_FILTERS );

	png_text title_text;
	title_text.compression = PNG_TEXT_COMPRESSION_NONE;
	title_text.key = "Title";
	title_text.text = "I can't believe it's not an image";
	png_set_text( png_ptr, info_ptr, &title_text, 1 );

	png_bytep *rowptrs = (png_bytep *)malloc( src->h * sizeof(png_bytep *) );

	for( int i = 0; i < src->h; i++ )
		rowptrs[i] = (png_byte *)&src->data[src->w * (flip ? src->h-1-i : i)];

	png_write_info( png_ptr, info_ptr );
	png_write_image( png_ptr, rowptrs );
	png_write_end( png_ptr, NULL );

	fclose( fp );

	png_destroy_write_struct( &png_ptr, &info_ptr );

	free( rowptrs );

	return true;
}


static void text_draw( bmp32 *dstbuf, int fontid, int dx, int dy, const char *text, const uint32_t col )
{
	uint32_t *dst = dstbuf->data;
	int height = dstbuf->h;
	int stride = dstbuf->w;

	const FontHeader *fh = font_GetInfo( fontid );
	FontGlyph *glyphs = fh->glyphs;

	int xx = 0;

	uint32_t cb0 = (col>>16)&0xff;
	uint32_t cg0 = (col>> 8)&0xff;
	uint32_t cr0 = (col    )&0xff;

	for( int i = 0; text[i]; i++ ) {
		uint8_t sym = text[i];
		FontGlyph *gl = &glyphs[sym];

		uint8_t *src = fh->bits + gl->offset;

		if( i != 0 ) xx += gl->left;

		for( int y = gl->startrow; y < gl->startrow+gl->rows; y++ ) {

			if( dy+y >= height ) break;

			for( int x = 0; x < gl->width; x += 2 ) {

				int v0;
				uint8_t p0,p0n;
				uint32_t b0, g0, r0;
				uint32_t dstb0, dstg0, dstr0;
				uint32_t pix;
				uint8_t p0p1 = *src++;

				int offset = (dy+y)*stride+dx+x+xx;

				p0  = (p0p1&0xf0)+(p0p1>>4);
				p0n = 0xff-p0;

				pix = dst[offset];
				b0 = (pix>>16)&0xff;
				g0 = (pix>> 8)&0xff;
				r0 = (pix    )&0xff;

				v0 = b0 * p0n + cb0 * p0; dstb0 = (v0 + (v0>>8) + 0x80)>>8;
				v0 = g0 * p0n + cg0 * p0; dstg0 = (v0 + (v0>>8) + 0x80)>>8;
				v0 = r0 * p0n + cr0 * p0; dstr0 = (v0 + (v0>>8) + 0x80)>>8;

				if( dx+x+xx < dstbuf->w ) dst[offset] = 0xff000000|(dstb0<<16)|(dstg0<<8)|dstr0;

				p0  = (p0p1&0x0f)+(p0p1<<4);
				p0n = 0xff-p0;

				pix = dst[offset+1];
				b0 = (pix>>16)&0xff;
				g0 = (pix>> 8)&0xff;
				r0 = (pix    )&0xff;

				v0 = b0 * p0n + cb0 * p0; dstb0 = (v0 + (v0>>8) + 0x80)>>8;
				v0 = g0 * p0n + cg0 * p0; dstg0 = (v0 + (v0>>8) + 0x80)>>8;
				v0 = r0 * p0n + cr0 * p0; dstr0 = (v0 + (v0>>8) + 0x80)>>8;

				if( dx+x+xx+1 < dstbuf->w ) dst[offset+1] = 0xff000000|(dstb0<<16)|(dstg0<<8)|dstr0;

			}

		}

		xx += gl->xwidth - gl->left;

	}

}


static void text_draw_alpha( bmp32 *dstbuf, int fontid, int dx, int dy, const char *text, const uint32_t col )
{
	uint32_t *dst = dstbuf->data;
	int height = dstbuf->h;
	int stride = dstbuf->w;

	const FontHeader *fh = font_GetInfo( fontid );
	FontGlyph *glyphs = fh->glyphs;

	int xx = 0;

	uint32_t colrgb = col&0x00ffffff;

	for( int i = 0; text[i]; i++ ) {
		uint8_t sym = text[i];
		FontGlyph *gl = &glyphs[sym];

		uint8_t *src = fh->bits + gl->offset;

		if( i != 0 ) xx += gl->left;

		for( int y = gl->startrow; y < gl->startrow+gl->rows; y++ ) {

			if( dy+y >= height ) break;

			for( int x = 0; x < gl->width; x += 2 ) {

				int v0;
				uint8_t p0,p0n;
				uint32_t a0;
				uint32_t dsta0;
				uint32_t pix;
				uint8_t p0p1 = *src++;

				int offset = (dy+y)*stride+dx+x+xx;

				p0  = (p0p1&0xf0)+(p0p1>>4);
				p0n = 0xff-p0;

				pix = dst[offset];
				a0 = (pix>>24)&0xff;

				v0 = a0 * p0n + 0xff * p0; dsta0 = (v0 + (v0>>8) + 0x80)>>8;

				if( dx+x+xx < dstbuf->w ) dst[offset] = (dsta0<<24)|colrgb;

				p0  = (p0p1&0x0f)+(p0p1<<4);
				p0n = 0xff-p0;

				pix = dst[offset+1];
				a0 = (pix>>24)&0xff;

				v0 = a0 * p0n + 0xff * p0; dsta0 = (v0 + (v0>>8) + 0x80)>>8;

				if( dx+x+xx+1 < dstbuf->w ) dst[offset+1] = (dsta0<<24)|colrgb;

			}

		}

		xx += gl->xwidth - gl->left;

	}

}


void bmp32_DrawText( bmp32 *dst, int fontid, bool alpha, int x, int y, const char *text, uint32_t col )
{
	if( alpha )
		text_draw_alpha( dst, fontid, x, y, text, col );
	else
		text_draw( dst, fontid, x, y, text, col );
}

void bmp32_DrawTextRight(  bmp32 *dst, int fontid, bool alpha, int y, const char *text, uint32_t col )
{
	int len = font_GetLen( text, fontid );

	if( alpha )
		text_draw_alpha( dst, fontid, dst->w-len, y, text, col );
	else
		text_draw( dst, fontid, dst->w-len, y, text, col );
}

void bmp32_DrawTextCenter( bmp32 *dst, int fontid, bool alpha, int y, const char *text, uint32_t col )
{
	int len = font_GetLen( text, fontid );

	if( alpha )
		text_draw_alpha( dst, fontid, (dst->w-len)/2, y, text, col );
	else
		text_draw( dst, fontid, (dst->w-len)/2, y, text, col );
}
