// Based on "Noise 3D Fly Through" by revers
// https://www.shadertoy.com/view/4ddXW4
// License: Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License.

// The original source has been modified so it runs in 60 fps on the NVidia Jetson TX2.
// More information about changes here: https://www.ignorantus.com/pages/xmasdemo2017/

#version 310 es
//#pragma optionNV(unroll all)
//#pragma optionNV(fastmath on)
precision highp float;
layout(location = 0) in vec2 v_texCoord;
layout(location = 0) out vec4 outColor;

#define TABLELEN 256
#define TABLEMASK (TABLELEN-1)

layout(std140,binding = 0) uniform stuff
{
    float width, height;
    float frame;
    float colscale;

    vec4 noisetable[TABLELEN];
};

#define iGlobalTime (frame)

const float FOV = 0.4f;
const float MarchDumping = 0.9f;

const float Far = 38.925f;

const float precis = 0.03;

#define MAX_STEPS 96

const float CameraSpeed = 4.5099998f;
const float TunnelSmoothFactor = 2.0f;
const float TunnelRadius = 0.85660005f;
const float TunnelFreqA = 0.18003f;
const float TunnelFreqB = 0.25f;
const float TunnelAmpA = 3.6230998f;
const float TunnelAmpB = 2.4324f;
const float NoiseIsoline = 0.319f;
const float NoiseScale = 2.9980001f;
const vec3 Color = vec3(0.085f, 0.658f, 1.0f);

#define M_NONE -1.0
#define M_NOISE 1.0


#define PI 3.1415926525f

float noise( vec3 x00 )
{
    ivec3 p00 = ivec3( floor( x00 ) );

    vec3 f00 = fract( x00 );
    f00 = f00 * f00 * ( 3.0 - 2.0 * f00 );

    uint n00 = uint( abs( p00.x + p00.y * 157 + 113 * p00.z ) );
    vec4 h00 = noisetable[(n00+0u)&uint(TABLEMASK)];
    vec4 h01 = noisetable[(n00+1u)&uint(TABLEMASK)];

    h00    = mix( h00,    h01,    f00.x );
    h00.xy = mix( h00.xz, h00.yw, f00.y );
    h00.x  = mix( h00.x,  h00.y,  f00.z );

    return h00.x;
}

float fbm( vec3 p )
{
    float f;
    f  = 0.5000 * noise(p); p *= 2.01;
    f += 0.2500 * noise(p); p *= 2.02;
    f += 0.1250 * noise(p);
    return f;
}


// by iq. http://iquilezles.org/www/articles/smin/smin.htm
float smax( float a, float b, float k )
{
    float h = clamp(0.5 + 0.5 * (b - a) * 0.5f, 0.0, 1.0);
    return mix(a, b, h) + 2.0f * h * (1.0 - h);
}

// From "Subterranean Fly-Through" by Shane https://www.shadertoy.com/view/XlXXWj
vec2 path( float z )
{
    return vec2(TunnelAmpA * sin(z * TunnelFreqA), TunnelAmpB * cos(z * TunnelFreqB));
//    return vec2( 1.0f, 1.0f );
}

float map( vec3 p )
{
    float d = ( fbm( p / NoiseScale ) - NoiseIsoline ) * NoiseScale;

    float d2 = length( p.xy - path( p.z ) ) - TunnelRadius;

    d = smax(d, -d2, TunnelSmoothFactor);

    return d;
}

float castRay( vec3 ro, vec3 rd )
{
    float t = 0.0f;

    float md = MarchDumping;

    for( int i = 0; i < MAX_STEPS; i++ ) {

        float res = map( ro + rd * t );
        if( res < precis ) break;

        t += res * md;

        if( i > 24 ) md *= 1.025f;

        if( t > Far ) break;

    }

    return t;
}

vec3 calcNormal( vec3 pos )
{
    vec2 eps = vec2(0.001, 0.0);

    vec3 nor = vec3(map(pos + eps.xyy) - map(pos - eps.xyy),
                    map(pos + eps.yxy) - map(pos - eps.yxy),
                    map(pos + eps.yyx) - map(pos - eps.yyx));

    return normalize(nor);
}


vec3 render( vec3 ro, vec3 rd )
{
    vec3 col = vec3(0.0);

    float t = castRay(ro, rd);

    if( t > Far ) return col;

    vec3 pos = ro + t * rd;
    vec3 nor = calcNormal(pos);

    // material
    col = Color + sin(t * 0.8) * 0.3;
    col += 0.3 * sin(vec3(0.15, 0.02, 0.10) * iGlobalTime * 6.0);

    // lighitng
    float occ = 1.0f;
    vec3 lig = -rd;
    float amb = clamp(0.5 + 0.5 * nor.y, 0.0, 1.0);
    float dif = clamp(dot(nor, lig), 0.0, 1.0);

    float fre = pow(clamp(1.0 + dot(nor, rd), 0.0, 1.0), 2.0);

    vec3 ref = reflect(rd, nor);
    float spe = pow(clamp(dot(ref, lig), 0.0, 1.0), 100.0);

    vec3 brdf = vec3(0.0);
    brdf += 1.20 * dif * vec3(1.00, 0.90, 0.60);
    brdf += 1.20 * spe * vec3(1.00, 0.90, 0.60) * dif;

    // Additional specular lighting trick,
    // taken from "Wet stone" by TDM
    // https://www.shadertoy.com/view/ldSSzV
//    nor = normalize(nor - normalize(rd + .5)*.2);
    nor = normalize(nor - normalize(pos) * 0.2);
    ref = reflect(rd, nor);
    spe = pow(clamp(dot(ref, lig), 0.0, 1.0), 100.0);
    brdf += 2.20 * spe * vec3(1.00, 0.90, 0.60) * dif;

    brdf += 0.40 * amb * vec3(0.50, 0.70, 1.00);
    brdf += 0.40 * fre * vec3(1.00, 1.00, 1.00);

    col = col * brdf;

    col = col * exp(-0.005 * t * t);

    return vec3(clamp(col, 0.0, 1.0));
}

mat3 rotationZ( float a )
{
    float sa = sin(a);
    float ca = cos(a);

    return mat3(ca, sa, 0.0, -sa, ca, 0.0, 0.0, 0.0, 1.0);
}

void main()
{
    vec2 coord = 2.0 * v_texCoord - 1.0;
    coord.x *= width / height;
    coord *= FOV;

    float t = iGlobalTime * CameraSpeed + 4.0 * 60.0;
    vec3 ro = vec3(path(t), t);

    t += 0.5;
    vec3 target = vec3(path(t), t);
    vec3 dir = normalize(target - ro);
    vec3 up = vec3(-0.9309864, -0.33987653, 0.1332234) * rotationZ(iGlobalTime * 0.05);
    vec3 upOrtho = normalize(up - dot(dir, up) * dir);
    vec3 right = normalize(cross(dir, upOrtho));

    vec3 rd = normalize(dir + coord.x * right + coord.y * upOrtho);

    vec3 col = render(ro, rd);

    col = pow(col, vec3(0.4545));

    outColor = vec4(col, colscale);
}
