// Based on "Torus Thingy 1" by bal-khan
// https://www.shadertoy.com/view/llSyWD
// License: Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License.

// The original source has been modified so it runs in 60 fps on the NVidia Jetson TX2.
// More information about changes here: https://www.ignorantus.com/pages/xmasdemo2017/

#version 310 es
precision highp float;
layout(location = 0) in vec2 v_texCoord;
layout(location = 0) out vec4 outColor;

layout(std140,binding = 0) uniform stuff
{
    float width, height;
    float frame;
    float colscale;
};

#define iTime frame

vec3 ret_col = vec3( 0.0f ); // torus color
vec3 h = vec3( 0.0f );       // light amount

const vec3 teal   = vec3(0.3 , .7, .9);
const vec3 orange = vec3(0.95, .5, .1);

#define I_MAX  52
#define FAR    30.0f

#define PI 3.14159
#define TAU (PI*2.0f)

// Don't use long library functions, like atan().
// Nvcc will just replace the call and don't bother with rescheduling
// instructiions around or inside it. So the library atan2() is a
// stallfest.

// IEEE Signal Processing Magazine ( Volume: 30, Issue: 1, Jan. 2013 )
// Full Quadrant Approximations for the Arctangent Function
// http://ieeexplore.ieee.org/document/6375931/

#define f2u( x ) floatBitsToUint( x )
#define u2f( x ) uintBitsToFloat( x )

float satan2( float y, float x )
{
    uint sign_mask = 0x80000000u;
    float b = 0.596227f;

    // Extract the sign bits
    uint ux_s  = (sign_mask & f2u(x) )^sign_mask;
    uint uy_s  = (sign_mask & f2u(y) )^sign_mask;

    // Determine the quadrant offset
    float q = float( ( ~ux_s & uy_s ) >> 29 | ux_s >> 30 );

    // Calculate the arctangent in the first quadrant
    float bxy_a = abs( b * x * y );
    float num = bxy_a + y * y;
    float atan_1q =  num / ( x * x + bxy_a + num );

    // Translate it to the proper quadrant
    uint uatan_2q = ((ux_s ^ uy_s) | f2u(atan_1q));
    return (q + u2f(uatan_2q)) * PI/2.0f - PI;
}

vec2 modA( vec2 p, float count )
{
    float an = TAU/count;
    float a = satan2(p.y,p.x)+an*.5;
    a = mod(a, an)-an*.5;
    return vec2(cos(a),sin(a))*length(p);
}

// for known s2, guess that PI/4 is close enough
vec2 modAs2( vec2 p, float count, float s2 )
{
    float an = TAU/count;
    float a = s2+an*.5 + PI/4.0f;
    a = mod(a, an)-an*.5;
    return vec2(cos(a),sin(a))*length(p);
}

vec2 rotate( vec2 v, float angle )
{
    return vec2(cos(angle)*v.x+sin(angle)*v.y,-sin(angle)*v.x+cos(angle)*v.y);
}

float scene(vec3 p)
{
    p.xz = rotate( p.xz, 1.57f - 0.150f * iTime );
    p.yz = rotate( p.yz, 1.57f - 0.125f * iTime );
    vec3 op = p;

    // center spikes
    float s2 = satan2( p.x, p.z );
    float var = ( s2 + PI )/ TAU * 40.0f;
    p.xz = modAs2( p.xz, 40.0f, s2 );
    p.x -= 8.0f;

    // replace mod(), hmm
    p.xy = rotate( p.xy, iTime * 0.25f * ((int(var-0.5f)&1) == 1 ? -1. : 1.) );

    vec2 q = vec2(length(p.xy)-4.,p.z);

    float mind = length(q)-.05;

    // orange filler
    s2 = satan2( p.x, p.y );
    float vir = ( s2 + PI ) / TAU;
    var = vir * 30.0f;
    p.xy = modAs2( p.xy, 30.0f, s2 );
    p.x -= 4.0f;
    q = vec2( length( p.zx ) - 0.25f, p.y );

    float ming = length( q ) - 0.05f;

    mind = min( mind, ming );

    ret_col = mind == ming ? orange : teal;

    // outer part
    p.xz = rotate( p.xz, iTime * 1.5f * ((int(var)&1) == 1 ? -1. : 1.) * ((int(vir)&1) == 1 ? -1. : 1.) );

    p.xz = modA( p.xz, 20.0f );
    p.x -= 0.25f;
    q = vec2( length( p.xy ) - 0.1f, p.z );

    float mint = length( q ) - 0.02f;

    mind = min( mind, mint );

    if( mind == mint ) ret_col = vec3( 0.2f, 0.7f, 0.4f );

    // dodecahedron
    op.zx = rotate( op.zx, iTime * 0.5f );
    op.xz = modA( op.xz, 25.0f );
    op.x -= 6.0f;
    op  /= 1.732; //sqrt(3.)

    float b = 0.075f;

    // should reconsider the op5,b stuff
    op = abs( op );
    vec3 op5 = 0.5f * op;

    float minl;
    minl =            max( max( op.x+op5.y-b, op.y+op5.z-b ), op.z+op5.x-b )  ; b *= .95;
    minl = max(minl, -max( max( op.x+op5.y-b, op.y-op5.z-b ), op.z+op5.x-b ) );
    minl = max(minl, -max( max( op.x-op5.y-b, op.y+op5.z-b ), op.z+op5.x-b ) );
    minl = max(minl, -max( max( op.x+op5.y-b, op.y+op5.z-b ), op.z-op5.x-b ) );
    // end dodecahedron

    mind = min( mind, minl );

    if( mind == minl ) ret_col = vec3( 0.5f, 0.2f, 0.8f );

    h += 0.125f * vec3( 0.2f, 0.1f, 0.3f ) / ( pow( minl, 25.0f ) + 0.5f ); // dodecahedron fog

    return (mind);
}

#if 0
vec2 march(vec3 pos, vec3 dir)
{
    vec2 dist = vec2(0.0, 0.0);
    int i;
    for( i = 0; i < I_MAX+1; i++ ) {
        vec3 p = pos + dir * dist.y;
        dist.x = scene(p);
        dist.y += dist.x;
        float dinamyceps = -dist.x+(dist.y)*(1.0f/1500.0f);
        if( dist.y > FAR || dist.x < dinamyceps || log(dist.y*dist.y/dist.x*(1.0f/100000.0f)) > 0.0f ) break;
    }
    return vec2( float(i), dist.y );
}
#endif

// Will be a bit noisy, but quicker
vec2 march(vec3 pos, vec3 dir)
{
    vec2 dist = vec2(0.0, 0.0);

    float esc = 0.01f;

    int i;
    for( i = 0; i < I_MAX; i += 2 ) {
        dist.x = scene( pos + dir * dist.y ); if( dist.x < esc ) break; dist.y += dist.x; if( dist.y > FAR ) break;
        dist.x = scene( pos + dir * dist.y ); if( dist.x < esc ) break; dist.y += dist.x; if( dist.y > FAR ) break;
    }

    return vec2( float(i), dist.y );
}

vec3 camera( vec2 uv )
{
    float fov = 1.;
    vec3  forw  = vec3(0.0, 0.0, -1.0);
    vec3  right = vec3(1.0, 0.0, 0.0);
    vec3  up    = vec3(0.0, 1.0, 0.0);

    return (normalize((uv.x) * right + (uv.y) * up + fov * forw));
}

void main()
{
    vec2 iResolution = vec2( width, height );
    vec2 f = (1.0f - v_texCoord) * iResolution;

    float t  = iTime*.125;
    vec3 col = vec3(0., 0., 0.);
    vec2 R = iResolution.xy,
          uv  = vec2(f-R/2.) / R.y;
    vec3 dir = camera(uv);
    vec3 pos = vec3(.0+cos(t*5.), .0+sin(t*5.)*.5, .0);

    pos.z = 20.*exp(-t*5.)+10.+.5*sin(t*10.);
    vec2 inter = (march(pos, dir));

    if( inter.y <= FAR)
        col.xyz = ret_col*(1.-inter.x*.025);
    else
        col = vec3( 0.0f );

    col += h;
    col *= clamp((1.5-length(uv)), .5, 1.);
    col *= clamp((1.1-length(uv)), .0, 1.);

    outColor =  vec4( col, colscale );
}
