// GPURay 2.0: Raytracer fragment shader.
// Written by Nils Liaaen Corneliusen.
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
// Please refer to the article at https://www.ignorantus.com/gpuray/ for more information.
// More information about changes here: https://www.ignorantus.com/pages/xmasdemo2017/

#version 310 es
precision highp float;
layout(location = 0) in vec2 v_texCoord;
layout(location = 0) out vec4 outColor;

#define OBJNUM (4*32)
#define REFLECTIONS 3

#define FARDIST 256.0f

// must hardcode max, since uniform indeterminate not supported
#define MAXW 1920
#define MAXH 1080

#define BLOCKW 16
#define BLOCKH 16

#define MAXWR ((MAXW+BLOCKW-1)&~(BLOCKW-1))
#define MAXHR ((MAXH+BLOCKH-1)&~(BLOCKH-1))

#define BLOCKSX (MAXWR/BLOCKW)
#define BLOCKSY (MAXHR/BLOCKH)

#define BLOCK_ANY 0
#define BLOCK_BG  1
#define BLOCK_SPH 2

//#pragma optionNV (unroll all)

layout(std140,binding = 0) uniform stuff
{
    float width, height;
    float frame;
    float colscale;
    vec4 obj_pos[OBJNUM];
    vec4 obj_col[OBJNUM];
    vec4 plane_norm;
    vec4 plane_col;
    vec4 eye;
    vec4 light_pos;
    vec4 startpos;
    float ax;
    float ay;
    uvec4 probe[BLOCKSX*BLOCKSY/4/4];
};


void main()
{
    vec3 ray_dir = normalize( vec3( startpos.x + ax * v_texCoord.x,
                                    startpos.y + ay * v_texCoord.y,
                                    startpos.z ) );

    vec3 col = vec3( 0.0f );

    vec3 ray_pos = eye.xyz;

	// Find probe val...
    int blx = int(floor(v_texCoord.x*width /float(BLOCKW)));
    int bly = int(floor(v_texCoord.y*height/float(BLOCKH)));
    int offset = bly*BLOCKSX+blx;
    uint val = probe[offset/4/4][(offset/4)&0x03];
    int prval = int((val>>((offset&0x03)<<3))&uint(0xff));

    if( v_texCoord.x < 1.0f - colscale ) {

        vec3 col = vec3( 0.0f, 0.0f, 0.0f );

        if( prval == BLOCK_SPH )
            col = vec3( 1.0f, 0.0f, 0.0f );
        else if( prval == BLOCK_BG )
            col = vec3( 0.5f, 0.5f, 0.5f );
        else if( prval == BLOCK_ANY )
            col = vec3( 0.0f, 0.0f, 1.0f );

        outColor = vec4( col,  colscale );
        return;

    }

    if( prval == BLOCK_SPH || prval == BLOCK_ANY ) {

        int refl;
        for( refl = 0; refl < REFLECTIONS; refl++ ) {

            float rt_dist = 0.0f;
            int rt_i = -1;
            vec3 rt_t0;

            vec4 v0, v1, v2, v3;

            // preload 0-3
            v0 = vec4( ray_pos - obj_pos[0].xyz, obj_pos[0].w );
            v1 = vec4( ray_pos - obj_pos[1].xyz, obj_pos[1].w );
            v2 = vec4( ray_pos - obj_pos[2].xyz, obj_pos[2].w );
            v3 = vec4( ray_pos - obj_pos[3].xyz, obj_pos[3].w );

            // trace_ray()
            for( int i = 0; i < OBJNUM; i += 32 ) {
                vec4 v4, v5, v6, v7;
                vec4 b0, d0;

                // preload 4-7
                v4 = vec4( ray_pos - obj_pos[i+4].xyz, obj_pos[i+4].w );
                v5 = vec4( ray_pos - obj_pos[i+5].xyz, obj_pos[i+5].w );
                v6 = vec4( ray_pos - obj_pos[i+6].xyz, obj_pos[i+6].w );
                v7 = vec4( ray_pos - obj_pos[i+7].xyz, obj_pos[i+7].w );

                // calc 0-3
                b0 = vec4( dot( v0.xyz, ray_dir ),
                           dot( v1.xyz, ray_dir ),
                           dot( v2.xyz, ray_dir ),
                           dot( v3.xyz, ray_dir ) );

                d0 = vec4( b0.x * b0.x - dot( v0.xyz, v0.xyz ) + v0.w,
                           b0.y * b0.y - dot( v1.xyz, v1.xyz ) + v1.w,
                           b0.z * b0.z - dot( v2.xyz, v2.xyz ) + v2.w,
                           b0.w * b0.w - dot( v3.xyz, v3.xyz ) + v3.w );

                if( any( greaterThan( d0, vec4( 0.0f ) ) ) ) {

                    vec4  t0 = 1.0f / (-b0 - sqrt( d0 ));
                    float mv = max( max( t0.x, t0.y ), max( t0.z, t0.w ) );

                    if( mv > rt_dist ) {
                        rt_dist = mv;
                        rt_t0   = t0.yzw;
                        rt_i    = i + 0;
                    }

                }

                // preload 8-11
                v0 = vec4( ray_pos - obj_pos[i+ 8].xyz, obj_pos[i+ 8].w );
                v1 = vec4( ray_pos - obj_pos[i+ 9].xyz, obj_pos[i+ 9].w );
                v2 = vec4( ray_pos - obj_pos[i+10].xyz, obj_pos[i+10].w );
                v3 = vec4( ray_pos - obj_pos[i+11].xyz, obj_pos[i+11].w );

                // calc 4-7
                b0 = vec4( dot( v4.xyz, ray_dir ),
                           dot( v5.xyz, ray_dir ),
                           dot( v6.xyz, ray_dir ),
                           dot( v7.xyz, ray_dir ) );

                d0 = vec4( b0.x * b0.x - dot( v4.xyz, v4.xyz ) + v4.w,
                           b0.y * b0.y - dot( v5.xyz, v5.xyz ) + v5.w,
                           b0.z * b0.z - dot( v6.xyz, v6.xyz ) + v6.w,
                           b0.w * b0.w - dot( v7.xyz, v7.xyz ) + v7.w );

                if( any( greaterThan( d0, vec4( 0.0f ) ) ) ) {

                    vec4  t0 = 1.0f / (-b0 - sqrt( d0 ));
                    float mv = max( max( t0.x, t0.y ), max( t0.z, t0.w ) );

                    if( mv > rt_dist ) {
                        rt_dist = mv;
                        rt_t0   = t0.yzw;
                        rt_i    = i + 4;
                    }

                }

                // preload 12-15
                v4 = vec4( ray_pos - obj_pos[i+12].xyz, obj_pos[i+12].w );
                v5 = vec4( ray_pos - obj_pos[i+13].xyz, obj_pos[i+13].w );
                v6 = vec4( ray_pos - obj_pos[i+14].xyz, obj_pos[i+14].w );
                v7 = vec4( ray_pos - obj_pos[i+15].xyz, obj_pos[i+15].w );

                // calc 8-11
                b0 = vec4( dot( v0.xyz, ray_dir ),
                           dot( v1.xyz, ray_dir ),
                           dot( v2.xyz, ray_dir ),
                           dot( v3.xyz, ray_dir ) );

                d0 = vec4( b0.x * b0.x - dot( v0.xyz, v0.xyz ) + v0.w,
                           b0.y * b0.y - dot( v1.xyz, v1.xyz ) + v1.w,
                           b0.z * b0.z - dot( v2.xyz, v2.xyz ) + v2.w,
                           b0.w * b0.w - dot( v3.xyz, v3.xyz ) + v3.w );

                if( any( greaterThan( d0, vec4( 0.0f ) ) ) ) {

                    vec4  t0 = 1.0f / (-b0 - sqrt( d0 ));
                    float mv = max( max( t0.x, t0.y ), max( t0.z, t0.w ) );

                    if( mv > rt_dist ) {
                        rt_dist = mv;
                        rt_t0   = t0.yzw;
                        rt_i    = i + 8;
                    }

                }

                // preload 16-19
                v0 = vec4( ray_pos - obj_pos[i+16].xyz, obj_pos[i+16].w );
                v1 = vec4( ray_pos - obj_pos[i+17].xyz, obj_pos[i+17].w );
                v2 = vec4( ray_pos - obj_pos[i+18].xyz, obj_pos[i+18].w );
                v3 = vec4( ray_pos - obj_pos[i+19].xyz, obj_pos[i+19].w );

                // calc 12-15
                b0 = vec4( dot( v4.xyz, ray_dir ),
                           dot( v5.xyz, ray_dir ),
                           dot( v6.xyz, ray_dir ),
                           dot( v7.xyz, ray_dir ) );

                d0 = vec4( b0.x * b0.x - dot( v4.xyz, v4.xyz ) + v4.w,
                           b0.y * b0.y - dot( v5.xyz, v5.xyz ) + v5.w,
                           b0.z * b0.z - dot( v6.xyz, v6.xyz ) + v6.w,
                           b0.w * b0.w - dot( v7.xyz, v7.xyz ) + v7.w );

                if( any( greaterThan( d0, vec4( 0.0f ) ) ) ) {

                    vec4  t0 = 1.0f / (-b0 - sqrt( d0 ));
                    float mv = max( max( t0.x, t0.y ), max( t0.z, t0.w ) );

                    if( mv > rt_dist ) {
                        rt_dist = mv;
                        rt_t0   = t0.yzw;
                        rt_i    = i + 12;
                    }

                }

                // preload 20-23
                v4 = vec4( ray_pos - obj_pos[i+20].xyz, obj_pos[i+20].w );
                v5 = vec4( ray_pos - obj_pos[i+21].xyz, obj_pos[i+21].w );
                v6 = vec4( ray_pos - obj_pos[i+22].xyz, obj_pos[i+22].w );
                v7 = vec4( ray_pos - obj_pos[i+23].xyz, obj_pos[i+23].w );

                // calc 16-19
                b0 = vec4( dot( v0.xyz, ray_dir ),
                           dot( v1.xyz, ray_dir ),
                           dot( v2.xyz, ray_dir ),
                           dot( v3.xyz, ray_dir ) );

                d0 = vec4( b0.x * b0.x - dot( v0.xyz, v0.xyz ) + v0.w,
                           b0.y * b0.y - dot( v1.xyz, v1.xyz ) + v1.w,
                           b0.z * b0.z - dot( v2.xyz, v2.xyz ) + v2.w,
                           b0.w * b0.w - dot( v3.xyz, v3.xyz ) + v3.w );

                if( any( greaterThan( d0, vec4( 0.0f ) ) ) ) {

                    vec4  t0 = 1.0f / (-b0 - sqrt( d0 ));
                    float mv = max( max( t0.x, t0.y ), max( t0.z, t0.w ) );

                    if( mv > rt_dist ) {
                        rt_dist = mv;
                        rt_t0   = t0.yzw;
                        rt_i    = i + 16;
                    }

                }

                // preload 24-27
                v0 = vec4( ray_pos - obj_pos[i+24].xyz, obj_pos[i+24].w );
                v1 = vec4( ray_pos - obj_pos[i+25].xyz, obj_pos[i+25].w );
                v2 = vec4( ray_pos - obj_pos[i+26].xyz, obj_pos[i+26].w );
                v3 = vec4( ray_pos - obj_pos[i+27].xyz, obj_pos[i+27].w );

                // calc 20-23
                b0 = vec4( dot( v4.xyz, ray_dir ),
                           dot( v5.xyz, ray_dir ),
                           dot( v6.xyz, ray_dir ),
                           dot( v7.xyz, ray_dir ) );

                d0 = vec4( b0.x * b0.x - dot( v4.xyz, v4.xyz ) + v4.w,
                           b0.y * b0.y - dot( v5.xyz, v5.xyz ) + v5.w,
                           b0.z * b0.z - dot( v6.xyz, v6.xyz ) + v6.w,
                           b0.w * b0.w - dot( v7.xyz, v7.xyz ) + v7.w );

                if( any( greaterThan( d0, vec4( 0.0f ) ) ) ) {

                    vec4  t0 = 1.0f / (-b0 - sqrt( d0 ));
                    float mv = max( max( t0.x, t0.y ), max( t0.z, t0.w ) );

                    if( mv > rt_dist ) {
                        rt_dist = mv;
                        rt_t0   = t0.yzw;
                        rt_i    = i + 20;
                    }

                }

                // preload 28-31
                v4 = vec4( ray_pos - obj_pos[i+28].xyz, obj_pos[i+28].w );
                v5 = vec4( ray_pos - obj_pos[i+29].xyz, obj_pos[i+29].w );
                v6 = vec4( ray_pos - obj_pos[i+30].xyz, obj_pos[i+30].w );
                v7 = vec4( ray_pos - obj_pos[i+31].xyz, obj_pos[i+31].w );

                // calc 24-27
                b0 = vec4( dot( v0.xyz, ray_dir ),
                           dot( v1.xyz, ray_dir ),
                           dot( v2.xyz, ray_dir ),
                           dot( v3.xyz, ray_dir ) );

                d0 = vec4( b0.x * b0.x - dot( v0.xyz, v0.xyz ) + v0.w,
                           b0.y * b0.y - dot( v1.xyz, v1.xyz ) + v1.w,
                           b0.z * b0.z - dot( v2.xyz, v2.xyz ) + v2.w,
                           b0.w * b0.w - dot( v3.xyz, v3.xyz ) + v3.w );

                if( any( greaterThan( d0, vec4( 0.0f ) ) ) ) {

                    vec4  t0 = 1.0f / (-b0 - sqrt( d0 ));
                    float mv = max( max( t0.x, t0.y ), max( t0.z, t0.w ) );

                    if( mv > rt_dist ) {
                        rt_dist = mv;
                        rt_t0   = t0.yzw;
                        rt_i    = i + 24;
                    }

                }

                // preload 0-3 next iteration
                v0 = vec4( ray_pos - obj_pos[i+32].xyz, obj_pos[i+32].w );
                v1 = vec4( ray_pos - obj_pos[i+33].xyz, obj_pos[i+33].w );
                v2 = vec4( ray_pos - obj_pos[i+34].xyz, obj_pos[i+34].w );
                v3 = vec4( ray_pos - obj_pos[i+35].xyz, obj_pos[i+35].w );

                // calc 28-31
                b0 = vec4( dot( v4.xyz, ray_dir ),
                           dot( v5.xyz, ray_dir ),
                           dot( v6.xyz, ray_dir ),
                           dot( v7.xyz, ray_dir ) );

                d0 = vec4( b0.x * b0.x - dot( v4.xyz, v4.xyz ) + v4.w,
                           b0.y * b0.y - dot( v5.xyz, v5.xyz ) + v5.w,
                           b0.z * b0.z - dot( v6.xyz, v6.xyz ) + v6.w,
                           b0.w * b0.w - dot( v7.xyz, v7.xyz ) + v7.w );

                if( any( greaterThan( d0, vec4( 0.0f ) ) ) ) {

                    vec4  t0 = 1.0f / (-b0 - sqrt( d0 ));
                    float mv = max( max( t0.x, t0.y ), max( t0.z, t0.w ) );

                    if( mv > rt_dist ) {
                        rt_dist = mv;
                        rt_t0   = t0.yzw;
                        rt_i    = i + 28;
                    }

                }

                if( rt_dist > 0.33f ) break;
            }

            if( rt_i == -1 ) break;

            ivec3 iv = ivec3( equal( rt_t0, vec3( rt_dist ) ) );
            rt_i += iv.x + iv.y + iv.y + iv.z + iv.z + iv.z;

            // sphere hit + reflection
            ray_pos += ray_dir * (1.0f / rt_dist);
            vec3 n   = normalize( ray_pos - obj_pos[rt_i].xyz );
            vec3 l   = normalize( light_pos.xyz - ray_pos );

            float diffuse  = max( dot( n, l ), 0.0f );
            float specular = pow( max( dot( ray_dir, l - n * 2.0f * diffuse ), 0.0f ), 8.0f );

            col += vec3( specular ) + obj_col[rt_i].xyz * diffuse;
            ray_dir = reflect( ray_dir, n );

        }

        if( refl != 0 ) {
            outColor = vec4( col, colscale );
            return;
        }

    }

    // plane/sky
    float rt_dist = -(dot( ray_pos, plane_norm.xyz ) + plane_norm.w) / dot( ray_dir, plane_norm.xyz );

    if( rt_dist <= 0.0f ) {
        // plane miss, sky
        vec2 dxy = v_texCoord - 0.5f;
        float dv = 1.0f - sqrt( dot( dxy, dxy ) + (sin(fract(frame/2.0f)*2.0f*3.14f)+1.0f)*0.5f );
        outColor = vec4( plane_col.yzx * dv, colscale );
        return;
    }

    //  plane hit
    ray_pos += ray_dir * rt_dist;
    ray_dir  = normalize( light_pos.xyz - ray_pos );

    // check shadow
    vec4 v0, v1, v2, v3;

    // preload 0-3
    v0 = vec4( ray_pos - obj_pos[0].xyz, obj_pos[0].w );
    v1 = vec4( ray_pos - obj_pos[1].xyz, obj_pos[1].w );
    v2 = vec4( ray_pos - obj_pos[2].xyz, obj_pos[2].w );
    v3 = vec4( ray_pos - obj_pos[3].xyz, obj_pos[3].w );

    int i;
    for( i = 0; i < OBJNUM; i += 16 ) {
        vec4 v4, v5, v6, v7;
        vec4 d0, d1, d2, d3;
        vec4 b0;

        // preload 4-7
        v4 = vec4( ray_pos - obj_pos[i+4].xyz, obj_pos[i+4].w );
        v5 = vec4( ray_pos - obj_pos[i+5].xyz, obj_pos[i+5].w );
        v6 = vec4( ray_pos - obj_pos[i+6].xyz, obj_pos[i+6].w );
        v7 = vec4( ray_pos - obj_pos[i+7].xyz, obj_pos[i+7].w );

        // calc 0-3
        b0 = vec4( dot( v0.xyz, ray_dir ),
                   dot( v1.xyz, ray_dir ),
                   dot( v2.xyz, ray_dir ),
                   dot( v3.xyz, ray_dir ) );

        d0 = vec4( b0.x * b0.x - dot( v0.xyz, v0.xyz ) + v0.w,
                   b0.y * b0.y - dot( v1.xyz, v1.xyz ) + v1.w,
                   b0.z * b0.z - dot( v2.xyz, v2.xyz ) + v2.w,
                   b0.w * b0.w - dot( v3.xyz, v3.xyz ) + v3.w );

        // preload 8-11
        v0 = vec4( ray_pos - obj_pos[i+ 8].xyz, obj_pos[i+ 8].w );
        v1 = vec4( ray_pos - obj_pos[i+ 9].xyz, obj_pos[i+ 9].w );
        v2 = vec4( ray_pos - obj_pos[i+10].xyz, obj_pos[i+10].w );
        v3 = vec4( ray_pos - obj_pos[i+11].xyz, obj_pos[i+11].w );

        // calc 4-7
        b0 = vec4( dot( v4.xyz, ray_dir ),
                   dot( v5.xyz, ray_dir ),
                   dot( v6.xyz, ray_dir ),
                   dot( v7.xyz, ray_dir ) );

        d1 = vec4( b0.x * b0.x - dot( v4.xyz, v4.xyz ) + v4.w,
                   b0.y * b0.y - dot( v5.xyz, v5.xyz ) + v5.w,
                   b0.z * b0.z - dot( v6.xyz, v6.xyz ) + v6.w,
                   b0.w * b0.w - dot( v7.xyz, v7.xyz ) + v7.w );

        // preload 12-15
        v4 = vec4( ray_pos - obj_pos[i+12].xyz, obj_pos[i+12].w );
        v5 = vec4( ray_pos - obj_pos[i+13].xyz, obj_pos[i+13].w );
        v6 = vec4( ray_pos - obj_pos[i+14].xyz, obj_pos[i+14].w );
        v7 = vec4( ray_pos - obj_pos[i+15].xyz, obj_pos[i+15].w );

        // calc 8-11
        b0 = vec4( dot( v0.xyz, ray_dir ),
                   dot( v1.xyz, ray_dir ),
                   dot( v2.xyz, ray_dir ),
                   dot( v3.xyz, ray_dir ) );

        d2 = vec4( b0.x * b0.x - dot( v0.xyz, v0.xyz ) + v0.w,
                   b0.y * b0.y - dot( v1.xyz, v1.xyz ) + v1.w,
                   b0.z * b0.z - dot( v2.xyz, v2.xyz ) + v2.w,
                   b0.w * b0.w - dot( v3.xyz, v3.xyz ) + v3.w );

        // preload 16-19
        v0 = vec4( ray_pos - obj_pos[i+16].xyz, obj_pos[i+16].w );
        v1 = vec4( ray_pos - obj_pos[i+17].xyz, obj_pos[i+17].w );
        v2 = vec4( ray_pos - obj_pos[i+18].xyz, obj_pos[i+18].w );
        v3 = vec4( ray_pos - obj_pos[i+19].xyz, obj_pos[i+19].w );

        // calc 12-15
        b0 = vec4( dot( v4.xyz, ray_dir ),
                   dot( v5.xyz, ray_dir ),
                   dot( v6.xyz, ray_dir ),
                   dot( v7.xyz, ray_dir ) );

        d3 = vec4( b0.x * b0.x - dot( v4.xyz, v4.xyz ) + v4.w,
                   b0.y * b0.y - dot( v5.xyz, v5.xyz ) + v5.w,
                   b0.z * b0.z - dot( v6.xyz, v6.xyz ) + v6.w,
                   b0.w * b0.w - dot( v7.xyz, v7.xyz ) + v7.w );

        if( any( greaterThan( d0, vec4( 0.0f ) ) ) ||
            any( greaterThan( d1, vec4( 0.0f ) ) ) ||
            any( greaterThan( d2, vec4( 0.0f ) ) ) ||
            any( greaterThan( d3, vec4( 0.0f ) ) ) ) break;

    }

    float diffuse = i == OBJNUM ? dot( plane_norm.xyz, ray_dir ) * 1.2f : 0.0f;
    outColor = vec4( plane_col.xyz * diffuse, colscale );

}
