// Based on "Transparent Lattice" by Shane
// https://www.shadertoy.com/view/Xd3SDs
// License: Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License.

// The original source has been modified so it runs in 60 fps on the NVidia Jetson TX2.
// More information about changes here: https://www.ignorantus.com/pages/xmasdemo2017/

#version 310 es
precision highp float;
layout(location = 0) in vec2 v_texCoord;
layout(location = 0) out vec4 outColor;

layout(std140,binding = 0) uniform stuff
{
    float width, height;
    float frame;
    float colscale;
};

#define iTime (frame)


// Cheap vec3 to vec3 hash. Works well enough, but there are other ways.
vec3 hash33( vec3 p )
{
    float n = sin(dot(p, vec3(7, 157, 113)));
    return fract(vec3(2097152, 262144, 32768)*n);
}


// A fake noise looking field. Pretty interesting.

float map( vec3 p )
{
    p = (cos(p*.315*2.5 + sin(p.zxy*.875*2.5)));

    float n = length(p);

    p = sin(p*6. + cos(p.yzx*6.));

    return n - 1. - abs(p.x*p.y*p.z)*.05;
}


// Not big on accuracy, but lower on operations. Few distance function calls are important
// during volumetric passes.
/*
vec3 calcNormal( vec3 p, float d )
{
    const vec2 e = vec2(0.01, 0);
    return normalize(vec3(d - map(p - e.xyy), d - map(p - e.yxy), d - map(p - e.yyx)));
}
*/

vec3 calcNormal( vec3 p, float d )
{
    vec3 r0 = cos( p * .315 * 2.5 + sin( p.zxy * .875 * 2.5 ) ); p -= 0.01f;
    vec3 r1 = cos( p * .315 * 2.5 + sin( p.zxy * .875 * 2.5 ) );

    vec3 v0 = vec3( r1.x, r0.y, r0.z );
    vec3 v1 = vec3( r0.x, r1.y, r0.z );
    vec3 v2 = vec3( r0.x, r0.y, r1.z );

    float n0 = length( v0 );
    float n1 = length( v1 );
    float n2 = length( v2 );

    v0 = sin( v0 * 6. + cos( v0.yzx * 6. ) );
    v1 = sin( v0 * 6. + cos( v0.yzx * 6. ) );

    float f = - 1.0f - abs( v1.x * v1.y * v1.z ) * .05;

    n0 = n0 - 1.0f - abs( v0.x * v0.y * v0.z ) * .05;
    n1 = n1 + f;
    n2 = n2 + f;

    return normalize( vec3( d - n0, d - n1, d - n2 ) );
}

/*
// Tetrahedral normal, to save a couple of "map" calls. Courtesy of IQ. Unfortunately, still
// not fast enough in this particular instance.
vec3 calcNormal(in vec3 p){

    // Note the slightly increased sampling distance, to alleviate artifacts due to hit point inaccuracies.
    vec2 e = vec2(0.0025, -0.0025);
    return normalize(e.xyy * map(p + e.xyy) + e.yyx * map(p + e.yyx) + e.yxy * map(p + e.yxy) + e.x * map(p + e.xxx));
}
*/


void main( void )
{
    vec2 iResolution = vec2( width, height );

    vec2 fragCoord = (1.0f - v_texCoord) * iResolution;

    // Screen coordinates.
    vec2 uv = (fragCoord.xy - iResolution.xy*.5 )/iResolution.y;

    // Unit direction ray. The last term is one of many ways to fish-lens the camera.
    // For a regular view, set "rd.z" to something like "0.5."
    vec3 rd = normalize(vec3(uv, (1.-dot(uv, uv)*.5)*.5)); // Fish lens, for that 1337, but tryhardish, demo look. :)

    // There are a few ways to hide artifacts and inconsistencies. Making things go fast is one of them. :)
    // Ray origin, scene color, and surface postion vector.
    vec3 ro = vec3(0., 0., iTime*1.5);
    vec3 col=vec3(0);
    vec3 sp, sn, lp, ld, rnd;

    // Swivel the unit ray to look around the scene.
    // Compact 2D rotation matrix, courtesy of Shadertoy user, "Fabrice Neyret."
    vec2 a = sin(vec2(1.5707963, 0) + iTime*0.375);
    rd.xz = mat2(a, -a.y, a.x)*rd.xz;
    rd.xy = mat2(a, -a.y, a.x)*rd.xy;

    lp = vec3(0, 1, 4);
    lp.xz = mat2(a, -a.y, a.x)*lp.xz;
    lp.xy = mat2(a, -a.y, a.x)*lp.xy;
    lp += ro;

    // Unit ray jitter is another way to hide artifacts. It can also trick the viewer into believing
    // something hard core, like global illumination, is happening. :)
    //rd *= 0.99 + hash33(rd)*0.02;

    // Some more randomization, to be used for color based jittering inside the loop.
    rnd = hash33(rd+311.);

    // Ray distance, bail out layer number, surface distance and normalized accumulated distance.
    // Note the slight jittering to begin with. It alleviates the subtle banding.
    float t = length(rnd)*.2, d, aD;
    int layers = 0;

    // Light variables.
    float lDist, s, l;

    // Surface distance threshold. Smaller numbers gives a thinner membrane, but lessens detail...
    // hard to explain. It's easier to check it out for yourself.
    float thD = .0125; // + smoothstep(-0.2, 0.2, sin(iTime*0.75 - 3.14159*0.4))*0.025;

    // Only a few iterations seemed to be enough. Obviously, more looks better, but is slower.
    for(int i=0; i<40; i++) {

        // Break conditions. Anything that can help you bail early usually increases frame rate.
        if(layers>31 || t>16. || dot(col, vec3(.299, .587, .114)) > 1. ) break; // 31

        // Current ray postion. Slightly redundant here, but sometimes you may wish to reuse
        // it during the accumulation stage.
        sp = ro+rd*t;

        d = map(sp); // Distance to nearest point on the noise surface.

        // If we get within a certain distance of the surface, accumulate some surface values.
        // Values further away have less influence on the total.
        //
        // aD - Accumulated distance. You could smoothly interpolate it, if you wanted.
        //
        // 1/.(1. + t*t*0.1) - Basic distance attenuation. Feel free to substitute your own.

        // Normalized distance from the surface threshold value to our current isosurface value.
        aD = (thD-abs(d)*31./32.)/thD;

        // If we're within the surface threshold, accumulate some color.
        // Two "if" statements in a shader loop makes me nervous. I don't suspect there'll be any
        // problems, but if there are, let us know.
        if(aD>0.0) // 0.8 -> 40
        {

            // Add the accumulated surface distance value, along with some basic falloff using the
            // camera to light distance, "lDist." There's a bit of color jitter there, too.

            sn = calcNormal(sp, d)*sign(d);
//            sn = calcNormal( sp, d );
//            uvec3 intsn = floatBitsToUint( sn );
//            uvec3 signy = (intsn ^ floatBitsToUint( d  ))& 0x80000000u;
//            sn = uintBitsToFloat((intsn&0x7fffffffu) | signy);


            ld = (lp - sp); //vec3(.5773)
            lDist = max(length(ld), .001);
//            lDist = length(ld);
            ld /= lDist;
            s = pow(max(dot(reflect(-ld, sn), -rd), 0.), 8.);
            l = max(dot(ld, sn), 0.);

            //float c = dot(sin(sp*128. - cos(sp.yzx*64.)), vec3(.166))+.5;
            col += ((l + .1) + vec3(.5, .7, 1)*s)*aD/(1. + lDist*0.25 + lDist*lDist*0.05)*.2;
            // Failed experiment with color jitter to take out more banding.
            //col += ((l + .05 + fract(rnd + i*27.)*.1) + vec3(.5, .7, 1)*s)*aD/(1. + lDist*0.25 + lDist*lDist*0.05)*.2;

            // The layer number is worth noting. Accumulating more layers gives a bit more glow.
            // Lower layer numbers allow a quicker bailout. A lot of it is guess work.
            layers++;

        }

        // Kind of weird the way this works. I think not allowing the ray to hone in properly is
        // the very thing that gives an even spread of values. The figures are based on a bit
        // of knowledge versus trial and error. If you have a faster computer, feel free to tweak
        // them a bit.
        t += max(abs(d)*.75, thD*.25);

    }

    t = min(t, 16.);

    col = mix(col, vec3(0), 1.-exp(-0.025*t*t));////1.-exp(-0.01*t*t) 1.-1./(1. + t*t*.1)

    // Mixing the greytone color with a firey orange vignette. There's no meaning
    // behind it. I just thought the artsy greyscale was a little too artsy.
    uv = abs(fragCoord.xy/iResolution.xy - .5); // Wasteful, but the GPU can handle it.
    col = mix(col, vec3(min(col.x*1.5, 1.), pow(col.x, 2.5), pow(col.x, 12.)).yxz,
               min( dot(pow(uv, vec2(4.)), vec2(1))*8., 1.));

    //col = vec3(min(col.z*1.5, 1.), pow(col.z, 2.5), pow(col.z, 12.));

    // Mixing the vignette colors up a bit more.
    col = mix(col, col.zxy, dot(sin(rd*5.), vec3(.166)) + 0.166);

    // Presenting the color to the screen.
//    outColor = vec4( sqrt(clamp(col*2.0f, 0., 1.)), 1.0 );
    outColor = vec4( sqrt( col * 1.8f ), colscale );
}
