// Based on "Twofield" by w23
// https://www.shadertoy.com/view/Xs2XDV
// License: Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License.

// The original source has been modified so it runs in 60 fps on the NVidia Jetson TX2.
// More information about changes here: https://www.ignorantus.com/pages/xmasdemo2017/

#version 310 es
//#pragma optionNV(unroll all)
//#pragma optionNV(fastmath on)
precision highp float;
layout(location = 0) in vec2 v_texCoord;
layout(location = 0) out vec4 outColor;

#define TABLELEN 256
#define N 8

layout(std140,binding = 0) uniform stuff
{
    float width, height;
    float frame;
    float colscale;

    vec4 l1pos, l2pos, l3pos, l4pos;
    vec4 l1color, l2color, l3color, l4color;
    vec4 b1[N];
    vec4 b2[N];
    vec4 O;
};

#define Lmin (12.5f)
#define Lmax (40.0f)
#define STEPS 40

#define iTime frame

#define t (iTime * .3)

// Simplest Lambert+Phong material shading model parameters
struct mat_t {
    vec3 diffuse;
    vec4 specular;
};

// There are two separate distance fields, each consists of 8 metaballs and a plane.

vec2 t1t2( vec3 p )
{
    float v1 = 0.0f;
    float v2 = 0.0f;

    for( int i = 0; i < N; ++i ) {
        vec3 b1 = p - b1[i].xyz;
        vec3 b2 = p - b2[i].xyz;

        v1 += 5. / dot(b1, b1);
        v2 += 5. / dot(b2, b2);

    }

    float d1 = 12. - p.y; v1 += 3. / (d1*d1);
    float d2 = 12. + p.y; v2 += 3. / (d2*d2);

    return vec2( v1, v2 );
}


float world( vec3 p )
{
    vec2 t12 = t1t2( p );

    float w1 = 1.0f - t12.x + t12.y;
    float w2 = 1.0f + t12.x - t12.y;

    return min( w1, w2 );
}

// Material is picked based on which field's distance function is the smallest
mat_t material( vec3 p )
{
    mat_t m;

    vec2 t12 = t1t2( p );

    float w1 = 1.0f - t12.x + t12.y;
    float w2 = 1.0f + t12.x - t12.y;

    if (w1 < w2) {
        // red
        m.diffuse  = vec3( 0.92f, 0.027f, 0.027f );
        m.specular = vec4( 0.9440f, 0.3189f, 0.3189f, 1000.0f );
    } else {
        // blue
        m.diffuse  = vec3( 0.14f, 0.17f, 1.57f );
        m.specular = vec4( 0.5699f, 0.5850f, 1.2850f, 100.0f );
    }

    return m;
}


vec3 normal( vec3 p )
{
    vec2 e = vec2(.001,0.);

    return normalize( vec3( world( p + e.xyy ) - world( p - e.xyy ),
                            world( p + e.yxy ) - world( p - e.yxy ),
                            world( p + e.yyx ) - world( p - e.yyx ) ) );
}

// Raymarcher
float trace( vec3 D )
{
    // The usual "step along the ray" loop
    float Lp;
    float L = Lmin;

    for( int i = 0; i < STEPS; i += 2 ) {

        float d;
        d = world( O.xyz + D * L ); if( d < 0.01 * L || L >= Lmax ) break; Lp = L; L += d * 1.5;
        d = world( O.xyz + D * L ); if( d < 0.01 * L || L >= Lmax ) break; Lp = L; L += d * 1.5;

    }

    // Binary search for more exact intersection position
    // Needed to fix artifacts due to overstep
    if (L < Lmax ) {

        for( int i = 0; i < 5; i++ ) {
            float Lm = (Lp + L) * .5;
            if( world(O.xyz + D * Lm) < .01*Lm ) L = Lm; else Lp = Lm;
        }

    }

    return L;
}


// Compute shading color from a single light
vec3 enlight( vec3 p, vec3 v, vec3 n, mat_t m, vec3 lpos, vec3 lcolor )
{
    vec3 ldir = lpos - p; // direction from the point to the light source
    float ldist2 = dot(ldir, ldir); // square distance to the light source
    float ldist = sqrt(ldist2); // distance
    ldir /= ldist; // ldir is now normalized
//    float shadow = shadowtrace(p, ldir, ldist-.2); // whether the point is in shadow
    float shadow = 1.0f;
    return shadow * lcolor * (
        // Diffuse term (simple Lambert)
        m.diffuse * max(0., dot(n,ldir))
        // Specular term (Phong-ish)
      + m.specular.rgb * pow(max(0.,dot(normalize(ldir-v), n)), m.specular.w)
        // Specular energy conservation? (or something, it was 3 years ago :D)
        * (m.specular.w + 2.) * (m.specular.w + 4.) / (24. * (m.specular.w + pow(2., -m.specular.w/2.)))
    ) / ldist2;
}

// A ridiculous hack to display light sources
// Basic idea is to check whether the ray is "close enough" to the light source position in screen space
// Which is done by calculating the angle between the actual pixel ray and a vector from origin to the light source

vec3 lightball( vec3 lpos, vec3 lcolor, vec3 O, vec3 D, float L )
{
    vec3 ldir = lpos - O;

    if( dot( ldir, ldir ) > L*L ) return vec3( 0.0f );

    float lv = 2.07f - ( ( ( lpos.z / 10.0f + 1.0f ) / 2.0f ) + 1.0f );

    float pw = pow( max( 0.0f, dot( normalize( ldir ), D ) ), 20000.0f * lv );

    return ( normalize( lcolor ) + 1.0f ) * pw;
}


#define vt v_texCoord

void main( void )
{
    float xedge  = 256.0f/1920.0f;
    float yedge2 = 208.0f/1080.0f;
    float xedge2 = 416.0f/1920.0f;

    if(  (vt.x < xedge  || vt.x > 1.0f - xedge) ||
        ((vt.x < xedge2 || vt.x > 1.0f - xedge2) && vt.y > yedge2 && vt.y < 1.0f - yedge2 ) ) {
        outColor = vec4( 0.0f, 0.0f, 0.0f, colscale );
        return;
    }

    vec2 uv = v_texCoord * 2. - 1.;
    uv.x *= width / height;

    vec3 D = normalize(vec3(uv,-1.6));

//    D = normalize(mat3(rg.xyz, rgxfw.xyz, -fw.xyz) * D);

    // Raymarch
    float L = trace( D );

    vec3 color = vec3( 0.0f );

    if( L < Lmax ) {
        // If a hit is detected, then get the position
        vec3 p = O.xyz + D * L;
        // And normal
        vec3 n = normal(p);
        // And material
        mat_t m = material(p);

        // Ambient term (I don't think it is visible much)
        color = .001 * m.diffuse * n;

        // Shade for all lights
        color += enlight(p, D, n, m, l1pos.xyz, l1color.xyz);
        color += enlight(p, D, n, m, l2pos.xyz, l2color.xyz);
        color += enlight(p, D, n, m, l3pos.xyz, l3color.xyz);
        color += enlight(p, D, n, m, l4pos.xyz, l4color.xyz);

        // Fog-like hack-ish attenuation, so the only the center of the scene is visible
        color *= (1. - smoothstep(12., 18., length(p)));

    } else L = 100.0f;

    // Add light sources hack
    color += lightball(l1pos.xyz, l1color.xyz, O.xyz, D, L );
    color += lightball(l2pos.xyz, l2color.xyz, O.xyz, D, L );
    color += lightball(l3pos.xyz, l3color.xyz, O.xyz, D, L );
    color += lightball(l4pos.xyz, l4color.xyz, O.xyz, D, L );

    outColor = vec4( color,colscale );
}
