// The Xmas Demo 2020 Edition Control Code.
// Written by Nils Liaaen Corneliusen.
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
// Please refer to the article at https://www.ignorantus.com/pages/xmasdemo2020/ for more information.

// ffmpeg -framerate 60 -i c:\users\ncorn\Xmas2020Frames\pic%05d.png -i data\music_amp.wav -framerate 60 -c:v libx264 -preset slow -crf 18 -pix_fmt yuv420p -s 1920x1080 -movflags +faststart xmas2020.mp4

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdbool.h>
#include <assert.h>
#include <sys/types.h>
#include <math.h>
#include <string.h>
#include <errno.h>

#include <GL/glew.h>
#include <GL/gl.h>

#define MIKMOD_STATIC
#include <mikmod.h>

#include "glstuff.h"
#include "fontdefs.h"
#include "vec.h"
#include "raytracer.h"
#include "bmp.h"

static bool   savepics = false;
static char  *savepath = "pics";
static bmp32 *savebuf  = NULL; // will be allocated if savepics set

static bool nomodule  = false;
static MODULE *module = NULL;

static bool nointro = false;

static int frame_total = 0; // actual total

// Will be set by target specific code in postRender() if ESC is pressed
int quit = 0;

// PNG startup/end image from megamonitor.c
extern unsigned char megamonitor[];

// Music in XM format from madxmas.c
extern unsigned char madxmas[];

// Shaders from cshader.c
extern const char *shader_galdance;
extern const char *shader_glowcity;
extern const char *shader_tracer;
extern const char *shader_noise3d;
extern const char *shader_quat;
extern const char *shader_colorful;
extern const char *shader_seascape;
extern const char *shader_trans;
extern const char *shader_twofield;
extern const char *shader_torus;
extern const char *shader_scaler;

// Background texts from bgtext.c
extern const char **bgtexts[];

#define WIDTH  1920
#define HEIGHT 1080

// *** Raytracer (gpu_tracer defined in raytracer.h)
worldinfo world0;

// *** Noise 3D
#define TABLELEN 256
#define TABLEMASK (TABLELEN-1)

typedef struct {
	float width, height;
	float frame;
	float colscale;

	v4 noisetable[TABLELEN];
} gpu_noise3d;

// *** Julia Quaternions
typedef struct {
	float width, height;
	float frame;
	float colscale;

	v4 start_pos;
	v4 mu_pos;
	v4 obj_col;
	float rot_xz;
	float epsilon;
	float ax;
	float ay;
} gpu_quat;

#define MIX_TIME (2*60)
#define HOLD_TIME (8*60)
#define MU_TIME (HOLD_TIME+MIX_TIME)

static const v4 postable[] = {
	{  0.0f,    0.0f,     0.0f,     0.0f    }, // sphere
	{ -1.0f,    0.2f,     0.0f,     0.0f    }, // stick
	{ -0.125f, -0.256f,   0.847f,   0.0895f }, // doh
	{ -0.2f,    0.8f,     0.0f,     0.0f    },
	{ -0.450f, -0.447f,   0.181f,   0.306f  },
};

// *** Seascape
typedef struct {
	float width, height;
	float frame;
	float colscale;
	v4 ang;
	v4 ori;
	v4 m0, m1, m2;
} gpu_seascape;

// *** Twofield
#define NUM 8
typedef struct {
	float width, height;
	float frame;
	float colscale;
	v4 l1pos, l2pos, l3pos, l4pos;
	v4 l1color, l2color, l3color, l4color;
	v4 b1[NUM];
	v4 b2[NUM];
	v4 O;
} gpu_twofield;

// *** The rest of the effects uses this:
typedef struct {
	float width, height;
	float frame;
	float colscale;
} gpu_default;

static const char *vertSrc =
	"#version 310 es                            \n"
	"layout(location = 0) in  vec4 a_position;  \n"
	"layout(location = 1) in  vec2 a_texCoord;  \n"
	"layout(location = 0) out vec2 v_texCoord;  \n"
	"void main()                                \n"
	"{                                          \n"
	"    gl_Position = a_position;              \n"
	"    v_texCoord  = a_texCoord;              \n"
	"}                                          \n";

static const char *simpletexSrc =
	"#version 310 es                               \n"
	"precision highp float;                        \n"
	"layout(location = 0) in vec2 v_texCoord;      \n"
	"layout(location = 0) out vec4 outColor;       \n"
	"uniform sampler2D sampid;                     \n"
	"void main()                                   \n"
	"{                                             \n"
	"    outColor = texture( sampid, v_texCoord ); \n"
	"}                                             \n";

static const char *info2[] = {
	"A Corneliusen & Julin Production 2020",
	"Source code: www.ignorantus.com/xmasdemo",
	"Shader code from www.shadertoy.com",
	"Optimizations by N. L. Corneliusen",
	"Music: \"Mad Dubs Xmas\" by Sjur Julin",
};

static int infoctr2 = -1;
#define INFOCNT2 (sizeof(info2)/sizeof(info2[0]))

static GLuint texid;

static GLint  Coolsamplerloc;
static GLint  Simplesamplerloc;
static GLint  samplerloc;

static const GLfloat defvert[] = {
	-1.0f,  1.0f, 0.0f,  0.0f, 0.0f,
	-1.0f, -1.0f, 0.0f,  0.0f, 1.0f,
	 1.0f, -1.0f, 0.0f,  1.0f, 1.0f,
	 1.0f,  1.0f, 0.0f,  1.0f, 0.0f
};

static const GLushort indices[] = { 0, 1, 2, 0, 2, 3 };

static void buf_to_screen( bmp32 *src, const float *verts, const GLushort *inds )
{
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, src->w, src->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, src->data );
	glVertexAttribPointer( 0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat),  verts );
	glVertexAttribPointer( 1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat), &verts[3] );
	glEnableVertexAttribArray( 0 );
	glEnableVertexAttribArray( 1 );
	glActiveTexture( GL_TEXTURE0 );
	glBindTexture( GL_TEXTURE_2D, texid );
	glUniform1i( samplerloc, 0 );
	glDrawElements( GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, inds );
}

static void draw_bm( bmp32 *src, int dstx, int dsty )
{
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, src->w, src->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, src->data );

	float vert[sizeof(defvert)/sizeof(float)];
	memcpy( vert, defvert, sizeof(defvert) );

	float ux = dstx / (float)(WIDTH-1);
	float uy = dsty / (float)(HEIGHT-1);
	float lx = (dstx + src->w - 1) / (float)(WIDTH-1);
	float ly = (dsty + src->h - 1) / (float)(HEIGHT-1);

	ux = ux * 2.0f - 1.0f;
	uy = uy * 2.0f - 1.0f;
	lx = lx * 2.0f - 1.0f;
	ly = ly * 2.0f - 1.0f;

	vert[ 0] =  ux; vert[ 1] = -uy;
	vert[ 5] =  ux; vert[ 6] = -ly;
	vert[10] =  lx; vert[11] = -ly;
	vert[15] =  lx; vert[16] = -uy;

	glVertexAttribPointer( 0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat),  vert );
	glVertexAttribPointer( 1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat), &vert[3] );
	glEnableVertexAttribArray( 0 );
	glEnableVertexAttribArray( 1 );

	glActiveTexture( GL_TEXTURE0 );
	glBindTexture( GL_TEXTURE_2D, texid );
	glUniform1i( samplerloc, 0 );
	glDrawElements ( GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, indices );
}

static void procFrame( void )
{
	MikMod_Update();

	postRender();

	if( savepics ) {
		char fname[160];
		glFinish();
		glReadPixels( 0, 0, WIDTH, HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, savebuf->data );
		snprintf( fname, sizeof(fname), "%s\\pic%05d.png", savepath, frame_total );
		printf( "Saving %s\n", fname );
		bmp32_WritePng( savebuf, fname, true );
	}

	frame_total++;
}

static void delay_buf( bmp32 *src, int cnt )
{
	for( int i = 0; i < cnt; i++ ) {
		buf_to_screen( src, defvert, indices );
		procFrame();
		if( quit ) break;
	}
}

#define CENTERX   720.0f
#define CENTERY   460.0f
#define RADX      640.0f
#define RADY      400.0f
#define CREDSPEED 360

static void gen_credits( bmp32 *dst, int ctr )
{
	const FontHeader *fh = font_GetInfo( 0 );
	bmp32_ClearAlpha( dst );

	int ictr = ctr % CREDSPEED;
	float fctr = (ictr / (float)CREDSPEED)*2.0f*PI;

	int xoff, yoff;

	xoff = (int)(CENTERX + sinf( fctr                       ) * RADX);
	yoff = (int)(CENTERY + cosf( fctr                       ) * RADY);

	bmp32_DrawText( dst, 1, false, xoff, yoff,                 "Optimizations by",    0x0000ff00 );
	bmp32_DrawText( dst, 1, false, xoff, yoff+fh->maxheight*2, "Nils L. Corneliusen", 0x0000ff00 );

	xoff = (int)(CENTERX + sinf( fctr + 1.0f*(2.0f*PI)/3.0f ) * RADX);
	yoff = (int)(CENTERY + cosf( fctr + 1.0f*(2.0f*PI)/3.0f ) * RADY);

	ictr = ctr % 480;
	bmp32_DrawText( dst, 1, false, xoff, yoff,                 "Music: \"Mad Dubs Xmas\" by",     0x00ffff00 );
	bmp32_DrawText( dst, 1, false, xoff, yoff+fh->maxheight*2, "              Sjur Julin", 0x00ffff00 );

	xoff = (int)(CENTERX + sinf( fctr + 2.0f*(2.0f*PI)/3.0f ) * RADX);
	yoff = (int)(CENTERY + cosf( fctr + 2.0f*(2.0f*PI)/3.0f ) * RADY);

	bmp32_DrawText( dst, 1, false, xoff, yoff,                 "Shader code from",  0x0000ffff );
	bmp32_DrawText( dst, 1, false, xoff, yoff+fh->maxheight*2, "www.shadertoy.com", 0x0000ffff );
}

#define LINES_PER_COL 35

static void gen_bg( bmp32 *dst, int id )
{
	const FontHeader *fh = font_GetInfo( 2 );
	const char **texts = bgtexts[id];

	bmp32_ClearAlpha( dst );

	int top = (HEIGHT - fh->maxheight * LINES_PER_COL)/2;

	for( int x = 0; x < 2; x++ ) {
		for( int y = 0; y < LINES_PER_COL; y++ ) {
			bmp32_DrawText( dst, 2, false, x*dst->w/2 + 16, top+fh->maxheight*y, *texts++, 0x00ffff00 );
		}
	}
}

bmp32 *copypic = NULL;

static void writetext( bmp32 *dst, int fontid, int x, int y, const char *text, uint32_t col )
{
	char buf[160];
	strcpy( buf, text );
	int len = (int)strlen( text );
	int xlen = font_GetLen( text, fontid );

	for( int i = 1; i <= len; i++ ) {
		// First copy in old bg
		for( int yy = 0; yy < font_GetInfo( fontid )->maxheight; yy++ ) {
			memcpy( dst->data+(y+yy)*dst->w + x, copypic->data+(y+yy)*dst->w + x, xlen*4 );
		}
		// Move termination
		char save = buf[i];
		buf[i] = 0;
		bmp32_DrawText( dst, fontid, false, x, y, buf, col );
		buf[i] = save;
		delay_buf( dst, 1 );
		if( quit ) break;
	}
}

static void writetextcenter( bmp32 *dst, int fontid, int y, const char *text, uint32_t col )
{
	writetext( dst, fontid, (dst->w - font_GetLen( text, fontid ) )/2, y, text, col );
}

typedef enum {
	DEMO_GALDANCE,
	DEMO_GLOWCITY,
	DEMO_TRACER,
	DEMO_NOISE3D,
	DEMO_QUAT,
	DEMO_COLORFUL,
	DEMO_SEASCAPE,
	DEMO_TRANS,
	DEMO_TWOFIELD,
	DEMO_TORUS,
	DEMO_COUNT
} demotype;

typedef struct {
	demotype type;
	int winw;
	int winh;
	char *shader;
	const char *memshader;
	char *name;
	char *longname;
	char *desc;
	char *info;
} demo;

#ifdef _WIN32ssss
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow) {
# else
int main( int argc, char *argv[] ) {
	(void)argc;
	(void)argv;
#endif

	const FontHeader *fh = font_GetInfo( 0 );

	const demo demos[DEMO_COUNT] = {
		{ DEMO_GALDANCE, 1920, 1080, "galdance.glsl", shader_galdance, "galdance", "Galactic Dance",             "Based on code by Sinuousity",   "It's a... thing."                                    },
		{ DEMO_GLOWCITY, 1280,  720, "glowcity.glsl", shader_glowcity, "glowcity", "Glow City",                  "Based on code by mhnewman",     "A noise field city and some... err, many beacons."   },
		{ DEMO_TRACER,   1920, 1080, "tracer.glsl",   shader_tracer,   "tracer",   "Raytracer",                  "By Nils L. Corneliusen",        "128 spheres and shadows. CPU preprocessing."         },
		{ DEMO_NOISE3D,  1024,  576, "noise3d.glsl",  shader_noise3d,  "noise3d",  "Noise 3D Fly Through",       "Based on code by revers",       "Tunnel with super-fast noise distortion."            },
		{ DEMO_QUAT,     1920, 1080, "quat.glsl",     shader_quat,     "quat",     "Julia Quaternions",          "Based on code by Keenan Crane", "Raytracing, rotating and morphing."                   },
		{ DEMO_COLORFUL, 1600,  900, "colorful.glsl", shader_colorful, "colorful", "More Colorful Than Average", "Based on code by ollj",         "A busload of sines and 1500 dot products per pixel." },
		{ DEMO_SEASCAPE, 1920, 1080, "seascape.glsl", shader_seascape, "seascape", "Seascape",                   "Based on code by TDM",          "Sea simulation is just 2 sines and 3 muls. Sea'ish." },
		{ DEMO_TRANS,    1280,  720, "trans.glsl",    shader_trans,    "trans",    "Transparent Blobs",          "Based on code by Shane",        "Feat. Nils' lucky surface effect."                   },
		{ DEMO_TWOFIELD, 1920, 1080, "twofield.glsl", shader_twofield, "twofield", "Twofield",                   "Based on code by w23",          "Two distance fields, four lights, eight spheres."    },
		{ DEMO_TORUS,    1024,  576, "torus.glsl",    shader_torus,    "torus",    "Torus Thingy",               "Based on code by bal-khan",     "Everything's a torus, except the dodecahedrons."     },
	};

	int demo_runtime       =  30*60;
	int demo_fadetime      =   3*60;
	int demo_zoomtime_all  =   3*60;
	int demo_zoomtime_mcta =   1*60;
	int demo_zoomtime      = 666; // will be set to either all or mcta in end of demo swap

	int demo_running = 0;

	const demo *currdemo = NULL;

	int textctr = 0;

	int demoframe  = demo_runtime; // starts from 0 every effect (0..1799)
	int demoframes = 0;            // starts 0 first effect, used for shader frame ctr (0..17999)

	int winwidth  = 0;
	int winheight = 0;

	bool ok = initOpenGL( true, WIDTH, HEIGHT );
	if( !ok ) return 1;

	if( !nomodule ) {

		MikMod_RegisterLoader( &load_xm );

#ifdef __linux__
		// Only OpenAL seems to be playing music without hanging randomly.
		MikMod_RegisterDriver( &drv_openal );
#else
		MikMod_RegisterAllDrivers();
#endif
		md_mode |= DMODE_INTERP;
		if( MikMod_Init( "" ) ) {
			printf( "MikMod_Init error: %s\n", MikMod_strerror( MikMod_errno ) );
			return 1;
		}

		module = Player_LoadMem( (char *)madxmas, 10000000, 16, 0 );
		if( module == NULL ) {
			printf( "MikMod module load error: %s\n", MikMod_strerror( MikMod_errno ) );
			return 1;
		}
	}

	GLuint cooltexProgram = loadGraphicsProgram( vertSrc, NULL, shader_scaler, "Texturer" );
	if( cooltexProgram == 0 ) return -1;

	GLuint simpletexProgram = loadGraphicsProgram( vertSrc, NULL, simpletexSrc, "SimpleTexturer" );
	if( simpletexProgram == 0 ) return -1;

	bmp32 *startpic = bmp32_ReadPngMem( megamonitor, 10000000, false );
	if( startpic == NULL ) return -1;

	// safe draw area on monitor
	int disp_ulx =  514;
	int disp_uly =  153;
//	int disp_brx = 1410;
	int disp_bry =  835;
//	int disp_w = disp_brx - disp_ulx + 1;
	int disp_h = disp_bry - disp_uly + 1;

	if( savepics )
		savebuf = bmp32_Alloc( WIDTH, HEIGHT );

//	bmp32_DrawRect( startpic, disp_ulx, disp_uly, disp_brx, disp_bry, 0x00000000 );

	bmp32 *donepic  = bmp32_Duplicate( startpic );

	copypic = bmp32_Duplicate( startpic );

	bmp32 *topbuf = bmp32_Alloc( WIDTH, fh->maxheight*1 );
	bmp32 *botbuf = bmp32_Alloc( WIDTH, fh->maxheight*2 );
	bmp32 *cenbuf = bmp32_Alloc(   832, fh->maxheight*3 );
	bmp32 *bgbuf  = bmp32_Alloc( WIDTH, HEIGHT );

	bmp32_DrawTextCenter( topbuf, 0, true, 0, "The Xmas Demo 2020 Edition", 0x00ffff00 );
	topbuf->h -= 6;
	topbuf->data += topbuf->w*6;

	glPixelStorei( GL_UNPACK_ALIGNMENT, 4 );
	glGenTextures( 1, &texid );
	glBindTexture( GL_TEXTURE_2D, texid );

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );

	Coolsamplerloc   = glGetUniformLocation( cooltexProgram,   "sampid" );
	Simplesamplerloc = glGetUniformLocation( simpletexProgram, "sampid" );

	GLint tex_scrw   = glGetUniformLocation( cooltexProgram,   "tex_scrw" );
	GLint tex_scrh   = glGetUniformLocation( cooltexProgram,   "tex_scrh" );

	glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );

	glViewport( 0, 0, WIDTH, HEIGHT );

	// *** Compile shaders
	GLuint demo_programs[DEMO_COUNT];

	bool success = true;

	for( int i = 0; i < DEMO_COUNT; i++ ) {

		const demo *de = &demos[i];

		printf( "Compiling %s... ", de->shader ); fflush( stdout );

		demo_programs[i] = loadGraphicsProgram( vertSrc, NULL, de->memshader, de->shader );

		if( demo_programs[i] == 0 ) {
			printf( "Fail\n" );
			success = false;
		} else {
			printf( "OK\n" );
		}

	}

	if( module ) Player_Start( module );

	bool cool = false;

	// *** Show intro
	if( !nointro ) {

		printf( "Frame: %5d Window: %4d*%4d Intro\n", frame_total, WIDTH, HEIGHT );

		glUseProgram( simpletexProgram );
		samplerloc = Simplesamplerloc;
		cool = false;

		int xoff = disp_ulx + 20 + 16;
		int ypos = disp_uly+(disp_h-fh->maxheight*19)/2;

        buf_to_screen( bgbuf, defvert, indices );

		writetextcenter( startpic, 1, ypos, "The Xmas Demo 2020 Edition",       0x0000ff00 ); ypos += fh->maxheight*2;
		writetextcenter( startpic, 0, ypos, "A Corneliusen & Julin Production", 0x0000ffff ); ypos += fh->maxheight*2;
		delay_buf( startpic, 25 ); if( quit ) return 1;

		writetextcenter( startpic, 0, ypos, "We have top men working on it right now...", 0x00ffffff ); ypos += fh->maxheight*2;
		delay_buf( startpic, 25 ); if( quit ) return 1;

		for( int i = 0; i < DEMO_COUNT; i++ ) {

			const demo *de = &demos[i];

			char buf[80];
			snprintf( buf, sizeof(buf), "Compiling %s... ", de->shader );
			int len = font_GetLen( buf, 0 );
			writetext( startpic, 0, xoff, ypos, buf, 0x00ffffff );

			delay_buf( startpic, 10 ); if( quit ) return 1;

			if( demo_programs[i] == 0 )
				bmp32_DrawText( startpic, 0, false, xoff+len, ypos, "Fail", 0x000000ff );
			else
				bmp32_DrawText( startpic, 0, false, xoff+len, ypos, "OK", 0x0000ff00 );

			ypos += fh->maxheight;

			delay_buf( startpic, 10 );
			if( quit ) return 1;

		}

		ypos += fh->maxheight;

		if( success )
			bmp32_DrawText( startpic, 0, false, xoff, ypos, "READY!", 0x0000ff00 );
		else
			bmp32_DrawText( startpic, 0, false, xoff, ypos, "Even in the future nothing works!", 0x000000ff );

		delay_buf( startpic, 85+8 ); if( !success || quit ) return 1;
	}

	int introframes = frame_total;

	// Gpu data blobs
	void *gpudata = NULL;
	int gpudatalen = sizeof(gpu_tracer); // largest

	gpu_default    *gpudef    = malloc( sizeof(gpu_default) );
	gpu_noise3d    *gpu3d     = malloc( sizeof(gpu_noise3d) );
	gpu_twofield   *gpu2f     = malloc( sizeof(gpu_twofield) );
	gpu_quat       *gpuquat   = malloc( sizeof(gpu_quat) );
	gpu_tracer     *gputracer = malloc( sizeof(gpu_tracer) );
	gpu_seascape   *gpusea    = malloc( sizeof(gpu_seascape) );

	gpu_default *gpucommon = NULL;


	// Initialize static demo stuff

	// Noise3D
	// Overlapping table: Reduces the entire noise calculation to two LDC.F32X4.
	for( int i = 0; i < TABLELEN; i++ ) {
		v4 v;
		v.x = fract( sinf( ((i +   0)&TABLEMASK) * 43758.5453123f ) );
		v.y = fract( sinf( ((i + 157)&TABLEMASK) * 43758.5453123f ) );
		v.z = fract( sinf( ((i + 113)&TABLEMASK) * 43758.5453123f ) );
		v.w = fract( sinf( ((i + 270)&TABLEMASK) * 43758.5453123f ) );
		gpu3d->noisetable[i] = v;
	}

	// Twofield
	gpu2f->l1color = v4_mul3( v4_set( 1.0f, 0.7f, 0.7f, 0.0f ), 40.0f );
	gpu2f->l2color = v4_mul3( v4_set( 0.7f, 0.7f, 1.0f, 0.0f ), 20.0f );
	gpu2f->l3color = v4_mul3( v4_set( 0.7f, 1.0f, 0.7f, 0.0f ), 60.0f );
	gpu2f->l4color = v4_set( 100.0f, 100.0f, 100.0f, 0.0f );

	// Quat
	gpuquat->ax  = 10.0f * WIDTH / HEIGHT;
	gpuquat->ay  = 10.0f;
	gpuquat->start_pos = v4_set( gpuquat->ax  * -0.5f - 0.0f,
								 gpuquat->ay  * -0.5f - 0.0f,
								 8.0f,
								 0.0f );
	gpuquat->rot_xz = 0.0f;
	gpuquat->epsilon = 0.0075f;

	// Raytracer
	world_init( &world0, gputracer, WIDTH, HEIGHT );


	GLuint grProgram = 0;

	GLuint sceneSSbo;

	initBuffer( &sceneSSbo, GL_UNIFORM_BUFFER, gpudatalen, NULL );

	glBindBufferBase( GL_UNIFORM_BUFFER, 0, sceneSSbo );

	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

	GLint old_fbo;
	GLuint framebuffer;

	glGenFramebuffers( 1, &framebuffer );

	// demos have negative Y up... great
	GLfloat vert[] = {
		-1.0f,  1.0f, 0.0f,  0.0f,  1.0f,
		-1.0f, -1.0f, 0.0f,  0.0f,  0.0f,
		 1.0f, -1.0f, 0.0f,  1.0f,  0.0f,
		 1.0f,  1.0f, 0.0f,  1.0f,  1.0f,
	};

	GLfloat demovert[sizeof(vert)/sizeof(float)];

	bool drawcen = false;

	int frameoffset = 0;

	while( !quit ) {

		if( demoframes % 420 == 0 ) {
			infoctr2++;
			if( infoctr2 == INFOCNT2 ) infoctr2 = 0;
		}

		// Clear crap
		glUseProgram( 0 );
		glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
		glClear( GL_COLOR_BUFFER_BIT );

		// Switch demo?
		if( demoframe == demo_runtime ) {

			if( demoframes != 0 ) demo_running++;

			if( demo_running == DEMO_COUNT ) {

				printf( "Frame: %5d Window: %4d*%4d End\n", frame_total, WIDTH, HEIGHT );

				// The end
				glUseProgram( simpletexProgram );
				samplerloc = Simplesamplerloc;
				cool = false;

				for( int i = 0; i <= 128; i++ ) {

                    buf_to_screen( bgbuf, defvert, indices );

					// fade
					for( int y = 0; y < donepic->h; y++ ) {
						uint32_t *row = donepic->data + y*donepic->w;
						for( int x = 0; x < donepic->w; x++ ) {
							uint32_t rgba = *row;
							*row++ = (rgba&0x00ffffff) | (i < 128 ? ((i*2)<<24) : (255<<24) );
						}
					}

					buf_to_screen( donepic, defvert, indices );
					procFrame();

				}

				int ypos = disp_uly+(disp_h-fh->maxheight*19)/2;

				writetextcenter( donepic, 1, ypos, "The Xmas Demo 2020 Edition",                 0x0000ff00 ); ypos += fh->maxheight*2; delay_buf( donepic, 30 );
				writetextcenter( donepic, 0, ypos, "A Corneliusen & Julin Production",           0x0000ffff ); ypos += fh->maxheight*2; delay_buf( donepic, 30 );
				writetextcenter( donepic, 0, ypos, "That's it, man. Game over, man. Game over!", 0x00ffffff ); ypos += fh->maxheight*8; delay_buf( donepic, 30 );
				writetextcenter( donepic, 0, ypos, "Source code:",                               0x00ffffff ); ypos += fh->maxheight*2; delay_buf( donepic, 30 );
				writetextcenter( donepic, 1, ypos, "www.ignorantus.com/xmasdemo",                0x00ffffff ); ypos += fh->maxheight*2; delay_buf( donepic, 30 );

				delay_buf( donepic, 300-18 );
				break;
			}

			grProgram = demo_programs[demo_running];

			currdemo = &demos[demo_running];

			winwidth  = currdemo->winw < WIDTH  ? currdemo->winw : WIDTH;
			winheight = currdemo->winh < HEIGHT ? currdemo->winh : HEIGHT;

			float xfac = winwidth  / (float)WIDTH;
			float yfac = winheight / (float)HEIGHT;

			printf( "Frame: %5d Window: %4d*%4d Demo %2d: %s\n", frame_total, winwidth, winheight, demo_running+1, currdemo->longname );

			// Resize and center demo window
			memcpy( demovert, vert, sizeof(vert) );

			demovert[ 0] *= xfac; demovert[ 5] *= xfac;
			demovert[10] *= xfac; demovert[15] *= xfac;
			demovert[ 1] *= yfac; demovert[ 6] *= yfac;
			demovert[11] *= yfac; demovert[16] *= yfac;

			switch( currdemo->type ) {
			case DEMO_NOISE3D:  gpudata = gpu3d;     gpudatalen = sizeof(gpu_noise3d);  break;
			case DEMO_TWOFIELD: gpudata = gpu2f;     gpudatalen = sizeof(gpu_twofield); break;
			case DEMO_QUAT:     gpudata = gpuquat;   gpudatalen = sizeof(gpu_quat);     break;
			case DEMO_TRACER:   gpudata = gputracer; gpudatalen = sizeof(gpu_tracer);   break;
			case DEMO_SEASCAPE: gpudata = gpusea;    gpudatalen = sizeof(gpu_seascape); break;
			default:            gpudata = gpudef;    gpudatalen = sizeof(gpu_default);  break;
			}
			gpucommon = (gpu_default *)gpudata;

			gpucommon->width    = (float)winwidth;
			gpucommon->height   = (float)winheight;
			gpucommon->colscale = 1.0f;

			demoframe = 0;

			// faster mcta zoom
			demo_zoomtime = currdemo->type == DEMO_COLORFUL ? demo_zoomtime_mcta : demo_zoomtime_all;

			// need this since shaders are unchanged from the 2017 version.
			// Makes MCTA look the same regardless of frame_demos.
			frameoffset = currdemo->type == DEMO_COLORFUL ? 9000 - demoframes : 0;
		}

		gpucommon->frame  = (demoframes+frameoffset)/60.0f;

		// Fade in/out
		if( demoframe <= demo_fadetime ) {
			// fade up <0.0..1.0]
			gpucommon->colscale = (demoframe)/(float)demo_fadetime;
		} else if( demoframe >= demo_runtime-demo_fadetime ) {
			// fade down <1.0..0.0]
			int fr = demo_fadetime - (demoframe-(demo_runtime-demo_fadetime));
			gpucommon->colscale = (fr-1)/(float)demo_fadetime;
		} else {
			gpucommon->colscale = 1.0f;
		}

		// Zoom in/out
		// Seriously, who writes this crap? Please go stand in the corner.
		int ww = winwidth, wh = winheight;
		if( demoframe <= demo_zoomtime ) {
			// zoom in width..winwidth
			// width..winwidth
			ww = winwidth  + (bgbuf->w - winwidth )*(demo_zoomtime-demoframe)/demo_zoomtime;
			wh = winheight + (bgbuf->h - winheight)*(demo_zoomtime-demoframe)/demo_zoomtime;
		} else if( demoframe >= demo_runtime-demo_zoomtime ) {
			// zoom out winwidth..width
			int f0 = demoframe - (demo_runtime - demo_zoomtime);
			ww = bgbuf->w - (bgbuf->w - winwidth )*(demo_zoomtime-f0)/demo_zoomtime;
			wh = bgbuf->h - (bgbuf->h - winheight)*(demo_zoomtime-f0)/demo_zoomtime;
		} else {
			ww = winwidth;
			wh = winheight;
		}

		// shrink window a bit, avoid edge artifacts
		if( ww != WIDTH  ) ww -= 2;
		if( wh != HEIGHT ) wh -= 2;

		if( currdemo->type != DEMO_GALDANCE && demoframe == demo_fadetime ) {
			// new bg texts after fade done
			gen_bg( bgbuf, textctr );
			textctr++;
		} else if( currdemo->type == DEMO_GALDANCE || (currdemo->type == DEMO_GLOWCITY && demoframe <= demo_fadetime-1) ) {
			// generate credits in galdance and the fade to glow city every frame
			if( demoframes > 84 ) { // really
				gen_credits( bgbuf, demoframes );
			}
		}

		// lower left info text
		bmp32_Clear( botbuf );
		bmp32_DrawText( botbuf, 0, true, 0, botbuf->h-fh->maxheight*1, currdemo->info, 0x0000ff00 );

		// lower center
		uint32_t infocol = (uint32_t)(gpucommon->colscale*255.0f);
		infocol = (infocol<<8)|infocol;

		bmp32_DrawTextCenter( botbuf, 0, true, botbuf->h-fh->maxheight*2, currdemo->longname, infocol );
		bmp32_DrawTextCenter( botbuf, 0, true, botbuf->h-fh->maxheight*1, currdemo->desc,     infocol );

		// lower right
		bmp32_DrawTextRight( botbuf, 0, true, botbuf->h-fh->maxheight*1, info2[infoctr2], 0x0000ff00 );

		// Run demo specific stuff each frame
		switch( currdemo->type ) {

		case DEMO_GALDANCE:
			if( demoframe > 84 && demoframe < demo_runtime-1 ) {
				if( drawcen == false ) {
					bmp32_Clear( cenbuf );
					bmp32_DrawTextCenter( cenbuf, 1, true, 0*fh->maxheight, "The Xmas Demo 2020 Edition",       0x00000000 );
					bmp32_DrawTextCenter( cenbuf, 0, true, 2*fh->maxheight, "A Corneliusen & Julin Production", 0x00000000 );
					drawcen = true;
				}
			} else {
				drawcen = false;
			}
			break;

		case DEMO_TORUS:
			if( demoframe >= 465 && demoframe <= 465+5*60 ) {
				if( drawcen == false ) {
					bmp32_Clear( cenbuf );
					bmp32_DrawTextCenter( cenbuf, 0, true, (cenbuf->h - fh->maxheight)/2, "SrDt 3454", 0x000000ff );
					drawcen = true;
				}
			} else {
				bmp32_Clear( cenbuf );
				drawcen = false;
			}
			break;

		case DEMO_TWOFIELD:
			// Calculate metaball trajectories
			for( int i = 0; i < NUM; ++i ) {
				float t = (demoframe-180)/60.0f * 0.3f;
				float fi = i*0.7f;
				gpu2f->b1[i] = v4_set( 3.7f*sinf(t     +fi), 1.0f+10.0f*cosf(t*1.1f+fi),  2.3f*sinf(t*2.3f+fi), 0.0f );
				fi = i*1.2f;
				gpu2f->b2[i] = v4_set( 4.4f*cosf(t*0.4f+fi),-1.0f-10.0f*cosf(t*0.7f+fi), -2.1f*sinf(t*1.3f+fi), 0.0f );
			}
			// Move lights
			float fw = gpucommon->frame*0.125f;
			float f0 = fract( fw + 0.25f );
			float f1 = fract( fw + 0.75f );
			float f2 = fract( fw + 0.5f  );
			float f3 = fract( fw + 0.0f  );
			gpu2f->l1pos = rotate( f0, 0.0f, 10.0f ); gpu2f->l1pos.y =  5.0f;
			gpu2f->l2pos = rotate( f1, 0.0f, 10.0f ); gpu2f->l2pos.y = -5.0f;
			gpu2f->l3pos = rotate( f2, 0.0f, 10.0f ); gpu2f->l3pos.y =  0.0f;
			gpu2f->l4pos = rotate( f3, 0.0f, 10.0f ); gpu2f->l4pos.y =  0.0f;
			gpu2f->O     = v4_set( 0.0f, 0.0f, 20.0f, 0.0f ); // rotation disabled
			break;

		case DEMO_QUAT:
		{
			int smu = demoframe/MU_TIME;
			int qfr = demoframe%MU_TIME;

			if( qfr < MIX_TIME ) {
				float wt = (qfr%MIX_TIME)/(float)(MIX_TIME);
				gpuquat->mu_pos = v4_mix( postable[smu], postable[smu+1], wt );
			} else {
				gpuquat->mu_pos = postable[smu+1];
			}

			gpuquat->obj_col = h2rgb( demoframe/520.0f );

			// rotate in xz plane
			float fr = demoframe / 240.0f;
			gpuquat->rot_xz = (fr-floorf(fr)) * 2.0f * PI;
			break;
		}

		case DEMO_SEASCAPE:
		{
			float time = gpucommon->frame * 0.3f;
			gpusea->ang = v4_set( sinf(time*3.0f)*0.1f, sinf(time)*0.2f+0.3f, time, 0.0f );
			gpusea->ori = v4_set( 0.0f, 3.5f , time*5.0f, 0.0f );
			v2 a1 = v2_set( sinf( gpusea->ang.x ), cosf( gpusea->ang.x ) );
			v2 a2 = v2_set( sinf( gpusea->ang.y ), cosf( gpusea->ang.y ) );
			v2 a3 = v2_set( sinf( gpusea->ang.z ), cosf( gpusea->ang.z ) );
			gpusea->m0 = v4_set(  a1.y*a3.y+a1.x*a2.x*a3.x, a1.y*a2.x*a3.x+a3.y*a1.x, -a2.y*a3.x, 0.0f );
			gpusea->m1 = v4_set( -a2.y*a1.x, a1.y*a2.y, a2.x, 0.0f );
			gpusea->m2 = v4_set(  a3.y*a1.x*a2.x+a1.y*a3.x, a1.x*a3.x-a1.y*a3.y*a2.x, a2.y*a3.y, 0.0f );
			break;
		}

		case DEMO_TRACER:
			world_timer( &world0, gputracer );
			world_probe( &world0, gputracer );
			break;

		default:
			break;
		}

		// *** Render demo to fbo
		glGetIntegerv( GL_FRAMEBUFFER_BINDING, &old_fbo );

		glUseProgram( grProgram );

		glBindFramebuffer( GL_FRAMEBUFFER, framebuffer );
		glBindTexture( GL_TEXTURE_2D, texid );

		void *src = gpucommon->colscale == 1.0f ? NULL : bgbuf->data;

		// the credits are drawn in bgbuf in galdance and the fade to glow city every frame, so render it
		if( currdemo->type == DEMO_GALDANCE || (currdemo->type == DEMO_GLOWCITY && demoframe <= demo_fadetime-1) ) {
			src = bgbuf->data;
		}

		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA,  WIDTH, HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL );
		if( src != NULL )
			glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, bgbuf->w, bgbuf->h, GL_RGBA, GL_UNSIGNED_BYTE, src );

		glFramebufferTexture2D( GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texid, 0 );

		copyToBuffer( GL_UNIFORM_BUFFER, sceneSSbo, gpudata, gpudatalen, GL_MAP_WRITE_BIT );

		glVertexAttribPointer( 0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat),  demovert );
		glVertexAttribPointer( 1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat), &demovert[3] );
		glEnableVertexAttribArray( 0 );
		glEnableVertexAttribArray( 1 );

		glDrawElements( GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, indices );

		// Scale fbo to display
		glBindFramebuffer( GL_FRAMEBUFFER, old_fbo );

		if( winwidth == WIDTH && winheight == HEIGHT ) {
			glUseProgram( simpletexProgram );
			samplerloc = Simplesamplerloc;
			cool = false;
		} else {
			glUseProgram( cooltexProgram );
			samplerloc = Coolsamplerloc;
			cool = true;
		}

		float scalevert[sizeof(defvert)/sizeof(float)];
		memcpy( scalevert, defvert, sizeof(defvert) );

		float sx   = (ww / (float)WIDTH);
		float sy   = (wh / (float)HEIGHT);
		float offx = (WIDTH -ww)*0.5f/WIDTH;
		float offy = (HEIGHT-wh)*0.5f/HEIGHT;

		scalevert[ 3] = offx;      scalevert[ 8] = offx;
		scalevert[13] = offx + sx; scalevert[18] = offx + sx;
		scalevert[ 4] = offy;      scalevert[19] = offy;
		scalevert[ 9] = offy + sy; scalevert[14] = offy + sy;

		glVertexAttribPointer( 0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat),  scalevert );
		glVertexAttribPointer( 1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat), &scalevert[3] );
		glEnableVertexAttribArray( 0 );
		glEnableVertexAttribArray( 1 );

		glActiveTexture( GL_TEXTURE0 );
		glBindTexture( GL_TEXTURE_2D, texid );

		glUniform1i( samplerloc, 0 );
		if( cool ) {
			glUniform1f( tex_scrw, (GLfloat)WIDTH );
			glUniform1f( tex_scrh, (GLfloat)HEIGHT );
		}

		glDrawElements( GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, indices );

		// Draw info windows
		glUseProgram( simpletexProgram );
		samplerloc = Simplesamplerloc;
		cool = false;

		if( currdemo->type != DEMO_GALDANCE ) draw_bm( topbuf, 0, 0 );

		draw_bm( botbuf, 0, HEIGHT-botbuf->h );

		if( drawcen ) draw_bm( cenbuf, (WIDTH-cenbuf->w)/2, (HEIGHT-cenbuf->h)/2 );

		procFrame();

		demoframe++;
		demoframes++;

	}

	int endframes = frame_total - demoframes - introframes;

	printf( "Total: %5d (Intro: %d Demos: %d End: %d)\n", frame_total, introframes, demoframes, endframes );

	printf( "Goodbye, Mr Bond\n" );

	if( module )
		MikMod_Exit();

	return 0;
}
